/*
 *  gcube
 *  Copyright (c) 2004 monk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *
 *      
 *         
 */

#include "diskio.h"
#include "mem.h"



__u32 load_dol (char *filename)
{
	FILE *dol;
	DOLHeader dh;
	int i;


	dol = fopen (filename, "r");
	if (!dol)
	{
		DEBUG (EVENT_EFATAL, ".dol: couldn't open %s", filename);
		return 0;
	}


	fread (&dh, sizeof (DOLHeader), 1, dol);
	
	for (i = 0; i < DOL_NTEXT; i++)
	{
		if (dh.text_offset[i])
		{
			fseek (dol, BSWAP32 (dh.text_offset[i]), SEEK_SET);
			fread (MEM_ADDRESS (BSWAP32 (dh.text_address[i])),
						 BSWAP32 (dh.text_size[i]), 1, dol);
			
			DEBUG (EVENT_INFO, ".dol: TEXT with size 0x%x loaded at 0x%x", BSWAP32 (dh.text_size[i]), BSWAP32 (dh.text_address[i]));
		}
	}
	
	for (i = 0; i < DOL_NDATA; i++)
	{
		if (dh.data_offset[i])
		{
			fseek (dol, BSWAP32 (dh.data_offset[i]), SEEK_SET);
			fread (MEM_ADDRESS (BSWAP32 (dh.data_address[i])),
						 BSWAP32 (dh.data_size[i]), 1, dol);

			DEBUG (EVENT_INFO, ".dol: DATA with size 0x%x loaded at 0x%x", BSWAP32 (dh.data_size[i]), BSWAP32 (dh.data_address[i]));
		}
	}
	
	fclose (dol);
	
	DEBUG (EVENT_INFO, ".dol: opened %s with entrypoint at 0x%x", filename, BSWAP32 (dh.entry_point));

	return BSWAP32 (dh.entry_point);
}


__u32 load_bin (char *filename)
{
	FILE *bin;
	int i;


	bin = fopen (filename, "r");
	if (!bin)
	{
		DEBUG (EVENT_EFATAL, ".bin: couldn't open %s", filename);
		return 0;
	}

	fseek (bin, 0, SEEK_END);
	i = ftell (bin);
	fseek (bin, 0, SEEK_SET);
	fread (MEM_ADDRESS (0x80003100), 1, i, bin);

	DEBUG (EVENT_INFO, ".bin: file %s, loaded 0x%x bytes at 0x80003100", filename, i);

	fclose (bin);

	return 0x80003100;
}


int is_elf (char *filename)
{
	FILE *f;
	Elf32_Ehdr hdr;


	f = fopen (filename, "rb");
	if (!f)
		return FALSE;

	fread (&hdr, sizeof (hdr), 1, f);
	if ((hdr.e_ident[EI_MAG0] == 0x7f  &&
			 hdr.e_ident[EI_MAG1] == 'E'   &&
			 hdr.e_ident[EI_MAG2] == 'L'   &&
			 hdr.e_ident[EI_MAG3] == 'F')  &&
			 hdr.e_ident[EI_CLASS] == ELFCLASS32)
	{
		if (BSWAP16(hdr.e_type) != ET_EXEC)
		{
			DEBUG (EVENT_EFATAL, ".elf: file %s is not executable (%d)", filename, hdr.e_type);
			return FALSE;
		}
		
		if (hdr.e_ident[EI_DATA] != ELFDATA2MSB)
		{
			DEBUG (EVENT_EFATAL, ".elf: only MSB encoding supported (%d)", hdr.e_ident[EI_DATA]);
			return FALSE;
		}

		return TRUE;
	}
	else
		return FALSE;
}


__u32 load_elf (char *filename)
{
	FILE *elf;
	Elf32_Ehdr hdr;
	Elf32_Phdr phdr;
	int i;


	if (!is_elf (filename))
		return FALSE;

	elf = fopen (filename, "rb");
	fread (&hdr, sizeof (hdr), 1, elf);

	hdr.e_phoff = BSWAP32 (hdr.e_phoff);
	hdr.e_phnum = BSWAP16 (hdr.e_phnum);
	for (i = 0; i < hdr.e_phnum; i++)
	{
		fseek (elf, hdr.e_phoff + i*sizeof (phdr), SEEK_SET);
		fread (&phdr, sizeof (phdr), 1, elf);

		if (BSWAP32 (phdr.p_type) == PT_LOAD && phdr.p_filesz != 0)
		{
			fseek (elf, BSWAP32 (phdr.p_offset), SEEK_SET);
			fread (MEM_ADDRESS (BSWAP32 (phdr.p_vaddr)),
						 BSWAP32 (phdr.p_filesz), 1, elf);

			DEBUG (EVENT_INFO, ".elf: section with size 0x%x loaded at 0x%x", BSWAP32 (phdr.p_filesz), BSWAP32 (phdr.p_vaddr));
		}
	}

	fclose (elf);

	DEBUG (EVENT_INFO, ".elf: opened %s with entrypoint at 0x%x", filename, BSWAP32 (hdr.e_entry));

	return BSWAP32 (hdr.e_entry);
}


int is_gcs (char *filename)
{
	FILE *f;
	GCSHeader hdr;


	f = fopen (filename, "rb");
	if (!f)
		return FALSE;

	fread (&hdr, sizeof (hdr), 1, f);
	fclose (f);

	if ((hdr.magic[0] == 0x7f) &&
			(hdr.magic[1] == 'G')  &&
			(hdr.magic[2] == 'C')  &&
			(hdr.magic[3] == 'S'))
		return TRUE;
	// for now assume that compressed file is a savestate
	else if ((hdr.magic[0] == 0x1f) && (hdr.magic[1] == 0x8B))
		return TRUE;
	else
		return FALSE;
}


void gcswrite (void *f, void *buff, unsigned int size, int compressed, int reverse)
{
	char *temp, *p = buff;
	int i;


	if (reverse)
	{
		temp = malloc (size);
		for (i = 0; i < size; i++)
			temp[i] = p[size - i - 1];
	}
	else
		temp = buff;

	if (compressed)
		gzwrite (f, temp, size);
	else
		fwrite (temp, 1, size, f);

	if (reverse)
		free (temp);
}


void gcsread (void *f, void *buff, unsigned int size, int reverse)
{
	char *temp, *p = buff;
	int i;


	if (reverse)
	{
		temp = malloc (size);

		gzread (f, temp, size);
		for (i = 0; i < size; i++)
			p[i] = temp[size - i - 1];
		
		free (temp);
	}
	else
		gzread (f, buff, size);
}


void save_state (char *filename, int compressed)
{
	FILE *f;
	gzFile gf = NULL;
	void *gcs;
	GCSHeader hdr;


	f = fopen (filename, "wb");
	if (!f)
	{
		DEBUG (EVENT_EMINOR, "..io: error writing savestate");
		return;
	}

	// todo: header shouldn't be compressed
	if (compressed)
	{
		int pos = ftell (f);
		
		gf = gzdopen (fileno (f), "wb");
		gzseek (gf, pos, SEEK_SET);

		gcs = gf;
	}
	else
		gcs = f;


	memset (&hdr, 0, sizeof (hdr));

	hdr.magic[0] = 0x7f;
	hdr.magic[1] = 'G';
	hdr.magic[2] = 'C';
	hdr.magic[3] = 'S';

	if (compressed)
		hdr.flags |= GCS_COMPRESSED_GZ;

	hdr.mem_size = MEM_SIZE;
	hdr.mem_offset = sizeof (hdr);
	hdr.regs_size = 4*4096;
	hdr.regs_offset = hdr.mem_offset + hdr.mem_size;
	hdr.fpregs_size = 8*32;
	hdr.fpregs_offset = hdr.regs_offset + hdr.regs_size;
	hdr.dspregs_size = RDSP_SIZE;
	hdr.dspregs_offset = hdr.fpregs_offset + hdr.fpregs_size;
	hdr.airegs_size = RAI_SIZE;
	hdr.airegs_offset = hdr.dspregs_offset + hdr.dspregs_size;
	hdr.cpregs_size = RCP_SIZE;
	hdr.cpregs_offset = hdr.airegs_offset + hdr.airegs_size;
	hdr.diregs_size = RDI_SIZE;
	hdr.diregs_offset = hdr.cpregs_offset + hdr.cpregs_size;
	hdr.eiregs_size = REXI_SIZE;
	hdr.eiregs_offset = hdr.diregs_offset + hdr.diregs_size;
	hdr.miregs_size = RMI_SIZE;
	hdr.miregs_offset = hdr.eiregs_offset + hdr.eiregs_size;
	hdr.peregs_size = RPE_SIZE;
	hdr.peregs_offset = hdr.miregs_offset + hdr.miregs_size;
	hdr.piregs_size = RPI_SIZE;
	hdr.piregs_offset = hdr.peregs_offset + hdr.peregs_size;
	hdr.siregs_size = RSI_SIZE;
	hdr.siregs_offset = hdr.piregs_offset + hdr.piregs_size;
	hdr.viregs_size = RVI_SIZE;
	hdr.viregs_offset = hdr.siregs_offset + hdr.siregs_size;

	gcswrite (gcs, &hdr, sizeof (hdr), compressed, FALSE);

	// memory
	gcswrite (gcs, MEM_ADDRESS (0), hdr.mem_size, compressed, FALSE);
	// cpu
	gcswrite (gcs, regs, hdr.regs_size, compressed, FALSE);
	gcswrite (gcs, fpregs, hdr.fpregs_size, compressed, FALSE);
	// hw
	gcswrite (gcs, rdsp, hdr.dspregs_size, compressed, TRUE);
	gcswrite (gcs, rai, hdr.airegs_size, compressed, TRUE);
	gcswrite (gcs, rcp, hdr.cpregs_size, compressed, FALSE);
	gcswrite (gcs, rdi, hdr.diregs_size, compressed, FALSE);
	gcswrite (gcs, rexi, hdr.eiregs_size, compressed, FALSE);
	gcswrite (gcs, rmi, hdr.miregs_size, compressed, FALSE);
	gcswrite (gcs, rpe, hdr.peregs_size, compressed, FALSE);
	gcswrite (gcs, rpi, hdr.piregs_size, compressed, FALSE);
	gcswrite (gcs, rsi, hdr.siregs_size, compressed, TRUE);
	gcswrite (gcs, rvi, hdr.viregs_size, compressed, TRUE);

	if (compressed)
		gzclose (gf);
	else
		fclose (f);

	DEBUG (EVENT_INFO, "..io: %ssavestate written to file %s",
				 (compressed) ? "compressed " : "", filename);
}


__u32 load_state (char *filename)
{
	FILE *f;
	gzFile gf;
	GCSHeader hdr;


	if (!is_gcs (filename))
		return 0;

	f = fopen (filename, "rb");
	gf = gzdopen (fileno (f), "rb");

	gcsread (gf, &hdr, sizeof (hdr), FALSE);
	
	// mem
	gzseek (gf, hdr.mem_offset, SEEK_SET);
	gcsread (gf, MEM_ADDRESS (0), hdr.mem_size, FALSE);
	// cpu
	gzseek (gf, hdr.regs_offset, SEEK_SET);
	gcsread (gf, regs, hdr.regs_size, FALSE);
	gzseek (gf, hdr.fpregs_offset, SEEK_SET);
	gcsread (gf, fpregs, hdr.fpregs_size, FALSE);
	// hw
	gzseek (gf, hdr.dspregs_offset, SEEK_SET);
	gcsread (gf, rdsp, hdr.dspregs_size, TRUE);
	gzseek (gf, hdr.airegs_offset, SEEK_SET);
	gcsread (gf, rai, hdr.airegs_size, TRUE);
	gzseek (gf, hdr.cpregs_offset, SEEK_SET);
	gcsread (gf, rcp, hdr.cpregs_size, FALSE);
	gzseek (gf, hdr.diregs_offset, SEEK_SET);
	gcsread (gf, rdi, hdr.diregs_size, FALSE);
	gzseek (gf, hdr.eiregs_offset, SEEK_SET);
	gcsread (gf, rexi, hdr.eiregs_size, FALSE);
	gzseek (gf, hdr.miregs_offset, SEEK_SET);
	gcsread (gf, rmi, hdr.miregs_size, FALSE);
	gzseek (gf, hdr.peregs_offset, SEEK_SET);
	gcsread (gf, rpe, hdr.peregs_size, FALSE);
	gzseek (gf, hdr.piregs_offset, SEEK_SET);
	gcsread (gf, rpi, hdr.piregs_size, FALSE);
	gzseek (gf, hdr.siregs_offset, SEEK_SET);
	gcsread (gf, rsi, hdr.siregs_size, TRUE);
	gzseek (gf, hdr.viregs_offset, SEEK_SET);
	gcsread (gf, rvi, hdr.viregs_size, TRUE);

	gzclose (gf);

	DEBUG (EVENT_INFO, "..io: savestate %s loaded", filename);
	return PC;
}


__u32 load_gc (char *filename)
{
	char ext[64];
	
	
	get_extension (ext, filename);
	if (0 == strcasecmp (ext, "bin"))
		return load_bin (filename);
	else if (0 == strcasecmp (ext, "dol"))
		return load_dol (filename);
	else if (is_elf (filename))
		return load_elf (filename);
	else if (is_gcs (filename))
		return load_state (filename);
	else
	{
		DEBUG (EVENT_EFATAL, "..io: unknown type of file %s", filename);
		return 0;
	}
}
