/*
 *  gcube
 *  Copyright (c) 2004 monk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *
 *      
 *         
 */

#include "gcube.h"



void (*vid_refresh) (void);
int gcube_running = FALSE;
char *quit_message = NULL;

#ifndef GDEBUG

int emu_state = GCUBE_RUN;

#else

static DebuggerState debugger;
# ifdef GCUBE_BY_DEFAULT
int emu_state = GCUBE_RUN;
# else
int emu_state = GDEBUG_RUN;
# endif
int gdebug_running = FALSE;


int command_find (char *cmd);
void debugger_parse_command (char *command);
void debugger_execute_command (void);
void debugger_cmd_step_into (char *args);
void debugger_cmd_detach (char *args);
void debugger_cmd_quit (char *args);
void debugger_cmd_help (char *args);
void debugger_map_add_item (__u32 address, char *s);

void mem_show (Window *win, int lines, unsigned int addr);
void code_show (Window *win, int lines, unsigned int addr);
void regs_show (Window *win, int sw);
void output_show (Window *win, OutputWin *out, int nlines, int pos);
void output_append (OutputWin *out, char *str);
void output_vappend (OutputWin *out, va_list ap, const char *format);
void output_appendf (OutputWin *out, char *format, ...);
void output_destroy (OutputWin *out);
void command_history_append (CommandWin *cmd, char *str);
void command_history_destroy (CommandWin *cmd);

void sprint_intmask (char *buff, __u32 b);


const char *reg_names[] =
{
	"R0", "R1", "R2", "R3", "R4", "R5", "R6", "R7", "R8", "R9", "R10", "R11",
	"R12", "R13", "R14", "R15", "R16", "R17", "R18", "R19", "R20", "R21", "R22",
	"R23", "R24", "R25", "R26", "R27", "R28", "R29", "R30", "R31",
	"SR0", "SR1", "SR2", "SR3", "SR4", "SR5", "SR6", "SR7", "SR8", "SR9", "SR10",
	"SR11", "SR12", "SR13", "SR14", "SR15",
	"CR", "MSR", "FPSCR", "TBL", "TBU", "PC", "IC", "SP", "XER", "LR", "CTR",
	"DEC", "SRR0", "SRR1", "HID0", "HID1", "HID2",
	"SDA1", "SDA2", "WPAR", "DMAU", "DMAL",
};

const int reg_indices[] =
{
	I_GPR+0, I_GPR+1, I_GPR+2, I_GPR+3, I_GPR+4, I_GPR+5, I_GPR+6, I_GPR+7, 
	I_GPR+8, I_GPR+9, I_GPR+10, I_GPR+11, I_GPR+12, I_GPR+13, I_GPR+14,
	I_GPR+15, I_GPR+16, I_GPR+17, I_GPR+18, I_GPR+19, I_GPR+20, I_GPR+21,
	I_GPR+22, I_GPR+23, I_GPR+24, I_GPR+25, I_GPR+26, I_GPR+27, I_GPR+28,
	I_GPR+29, I_GPR+30, I_GPR+31,
	I_SR+0, I_SR+1, I_SR+2, I_SR+3, I_SR+4, I_SR+5, I_SR+6, I_SR+7, I_SR+8,
	I_SR+9, I_SR+10, I_SR+11, I_SR+12, I_SR+13, I_SR+14, I_SR+15,
	I_CR, I_MSR, I_FPSCR, I_TBL, I_TBU, I_PC, I_IC, I_SP, I_XER, I_LR, I_CTR,
	I_DEC, I_SRR0, I_SRR1, I_HID0, I_HID1, I_HID2,
	I_SDA1, I_SDA2, I_WPAR, I_DMAU, I_DMAL,
};


UIDef constants[] =
{
	{ "on", 1 }, { "off", 0 },
	{ "true", 1}, { "false", 0 },
	{ "nop", 0x60000000 },
	{ "system_reset_exception", EXCEPTION_SYSTEM_RESET },
	{ "machine_check_exception", EXCEPTION_MACHINE_CHECK },
	{ "dsi_exception", EXCEPTION_DSI },
	{ "isi_exception", EXCEPTION_ISI },
	{ "external_exception", EXCEPTION_EXTERNAL },
	{ "alignment_exception", EXCEPTION_ALIGNMENT },
	{ "program_exception", EXCEPTION_PROGRAM },
	{ "fp_unavailable_exception", EXCEPTION_FP_UNAVAILABLE },
	{ "decrementer_exception", EXCEPTION_DECREMENTER },
	{ "system_call_exception", EXCEPTION_SYSTEM_CALL },
	{ "trace_exception", EXCEPTION_TRACE },
	{ "performance_monitor_exception", EXCEPTION_PERFORMANCE_MONITOR },
	{ "iabr_exception", EXCEPTION_IABR },
	{ "reserved_exception", EXCEPTION_RESERVED },
	{ "thermal_exception", EXCEPTION_THERMAL },
	{ "hsp_interrupt", INTERRUPT_HSP },
	{ "debug_interrupt", INTERRUPT_DEBUG },
	{ "cp_interrupt", INTERRUPT_CP },
	{ "pe_finish_interrupt", INTERRUPT_PE_FINISH },
	{ "pe_token_interrupt", INTERRUPT_PE_TOKEN },
	{ "vi_interrupt", INTERRUPT_VI },
	{ "mem_interrupt", INTERRUPT_MEM },
	{ "dsp_interrupt", INTERRUPT_DSP },
	{ "ai_interrupt", INTERRUPT_AI },
	{ "exi_interrupt", INTERRUPT_EXI },
	{ "si_interrupt", INTERRUPT_SI },
	{ "di_interrupt", INTERRUPT_DI },
	{ "rsw_interrupt", INTERRUPT_RSW },
	{ "error_interrupt", INTERRUPT_ERROR },
};


UIPDef variables[] =
{
	{ "INTSR", &INTSR },
	{ "INTMR", &INTMR },

	{ "code", &debugger.code_address },
	{ "mem", &debugger.mem_address },
	{ "last_pc", &debugger.last_pc },
	{ "refresh_rate", &debugger.refresh_rate },
};


unsigned int bin_to_dec (char *s)
{
	unsigned int i, len, res = 0;


	len = strlen (s);
	for (i = 0; i < len; i++)
		res += (s[i] - '0') * (1 << (len - i - 1));

	return res;
}


// returns real index (after wrapping)
int output_item_index (OutputWin *out, int n)
{
	int nn;


	nn = n + out->nfirst_item;
	if (nn >= out->nitems)
		nn -= out->nitems;
	
	return nn;
}


// returns number of characters extracted
int args_extract_uint (char *args, unsigned int *d)
{
	char buff[LINE_WIDTH + 1];
	int n = 0;


	// try hex
	if (1 <= sscanf (args, " 0x%x%n", d, &n))
		return n + 1;

	// try bin
	if (1 <= sscanf (args, " 0b%s%n", buff, &n))
	{
		*d = bin_to_dec (buff);
		return n + 1;
	}
	
	// try dec
	if (1 <= sscanf (args, " %d%n", d, &n))
		return n + 1;

	return 0;
}


int args_extract_str (char *args, char *d)
{
	int n = 0;


	if (1 <= sscanf (args, "%s%n", d, &n))
		return n + 1;

	return 0;
}


int args_extract_constant (char *args, unsigned int *c)
{
	char r[LINE_WIDTH + 1] = {0};
	int i, n;


	if ((n = args_extract_str (args, r)))
		for (i = 0; i < sizeof (constants) / sizeof (UIDef); i++)
			if (0 == strcasecmp (r, constants[i].s))
			{
				*c = constants[i].ui;
				return n;
			}

	return 0;
}


// gives the index of the variable
int args_extract_variable (char *args, unsigned int *index)
{
	char r[LINE_WIDTH + 1] = {0};
	int i, n;


	if ((n = args_extract_str (args, r)))
		for (i = 0; i < sizeof (variables) / sizeof (UIPDef); i++)
			if (0 == strcasecmp (r, variables[i].s))
			{
				*index = i;
				return n;
			}

	return 0;
}


// gives the index of the register (in the reg_names table)
int args_extract_reg (char *args, unsigned int *regindex)
{
	char r[LINE_WIDTH + 1] = {0};
	int i, n;


	if ((n = args_extract_str (args, r)))
		for (i = 0; i < sizeof (reg_names) / sizeof (char *); i++)
			if (0 == strcasecmp (r, reg_names[i]))
			{
				*regindex = i;
				return n;
			}

	return 0;
}


int args_extract_map_item (char *args, MapItem **mi)
{
	char r[LINE_WIDTH + 1] = {0};
	int n;
	MapItem *p;


	p = debugger.map;
	if ((n = args_extract_str (args, r)))
		while (p)
		{
			if (0 == strcasecmp (r, p->s))
			{
				*mi = p;
				return n;
			}
			else
				p = p->next;
		}

	return 0;
}


// try to parse the string and extract uint value
int args_parse_uint (char *args, unsigned int *d)
{
	unsigned int n;
	MapItem *mi;


	if ((n = args_extract_reg (args, d)))
	{
		*d = REGS (reg_indices[*d]);
		return n;
	}
	else if ((n = args_extract_constant (args, d)))
		return n;
	else if ((n = args_extract_variable (args, d)))
	{
		*d = *variables[*d].uip;
		return n;
	}
	else if ((n = args_extract_uint (args, d)))
		return n;
	else if ((n = args_extract_map_item (args, &mi)))
	{
		*d = mi->address;
		return n;
	}

	return 0;
}


int args_parse_subexpression (char *args, unsigned int *d)
{
	char buff[LINE_WIDTH + 1];
	char op[LINE_WIDTH + 1];
	int n = 0, k = 0, b = 0;


	sscanf (args, " %[^*/-+^%&|<> ]%n", buff, &n);
	if (!n || !args_parse_uint (buff, d))
		return 0;

	while (1)
	{
		k = 0;
		*op = '\0';
		sscanf (args + n, " %[*/-+^%&|<>]%n", op, &k);
		if (!k)
			return n;
		
		n += k;
		k = 0;
		sscanf (args + n, " %[^*/-+^%&|<> ]%n", buff, &k);
		if (!k || !args_parse_uint (buff, &b))
			return n;

		switch (*op)
		{
			case '*':
				*d *= b;
				break;
			
			case '/':
				*d /= b;
				break;

			case '^':
				*d ^= b;
				break;
		
			case '%':
				*d %= b;
				break;
		
			case '&':
				*d &= b;
				break;
		
			case '|':
				*d |= b;
				break;
		
			case '<':
				if (op[1] == '<')
					*d <<= b;
				break;
		
			case '>':
				if (op[1] == '>')
					*d >>= b;
				break;
		
			default:
				return n - 1;
		}

		n += k;
	}
}


int args_parse_expression_uint (char *args, unsigned int *d)
{
	int a = 0, n = 0, k = 0;


	*d = 0;
	if (!(n = args_parse_subexpression (args, d)))
		return 0;

	while (1)
	{
		if (!(k = args_parse_subexpression (args + n + 1, &a)))
			return n + 1;

		switch (args[n])
		{
			case '+':
				n++;
				*d += a;
				break;
			
			case '-':
				n++;
				*d -= a;
				break;

			default:
				return n + 1;
		}

		n += k;
	}
}


int breakpoint_reached (void)
{
	int i;


	for (i = 0; i < MAX_BREAKPOINTS; i++)
		if (debugger.breakpoint_enabled[i] && debugger.breakpoints[i] == PC)
			return i+1;

	return FALSE;
}


void debugger_cmd_run (char *args)
{
	int i = 0, b = 0;


	gdebug_running = TRUE;

	while (gdebug_running)
	{
		cpu_execute ();

		vid_refresh ();

		if (i++ > debugger.refresh_rate)
		{
			debugger.code_address = PC;
			screen_redraw (&debugger.screen);
			i = 0;
		}
		
		if ((b = breakpoint_reached ()))
		{
			output_appendf (&debugger.output, "breakpoint %d reached", b-1);
			break;
		}
	}

	debugger.code_address = PC;
}


// run the code until program conunter reaches next instruction
// used to execute functions
// under some conditions the next instruction might never be reached
void debugger_cmd_step_over (char *args)
{
	__u32 next = PC + 4;


	gdebug_running = TRUE;

	while (gdebug_running)
	{
		cpu_execute ();

		vid_refresh ();

		if (PC == next)
			break;
	}

	gdebug_running = FALSE;	
	debugger.code_address = PC;
}


void debugger_cmd_mem (char *args)
{
	__u32 n, address, data;


	if ((n = args_parse_expression_uint (args, &address)))
	{
		if (args_parse_expression_uint (args + n, &data))
		{
			switch (address % 4)
			{
				case 0:
					MEM_WWORD (address, data);
					break;
				
				case 2:
					MEM_WHALF (address, data);
					break;

				case 1:
				case 3:
					MEM_WBYTE (address, data);
					break;
			}
		}

		debugger.mem_address = address;
	}
	else
		output_append (&debugger.output, "* usage: mem address [data]");
}


void debugger_cmd_code (char *args)
{
	__u32 d;


	if (args_parse_expression_uint (args, &d))
		debugger.code_address = d;
	else
		output_append (&debugger.output, "* usage: code address");
}


void debugger_cmd_reload (char *args)
{
	PC = gcube_load_file (debugger.filename);

	if (PC == 0)
	{
		output_appendf (&debugger.output, "error: can't load file %s", debugger.filename);
		PC = debugger.code_address = debugger.starting_pc;
	}
	else
		debugger.code_address = debugger.starting_pc = PC;
}


void debugger_cmd_load (char *args)
{
	char buff[256];


	if (!args_extract_str (args, buff))
	{
		output_append (&debugger.output, "* usage: load file");
		return;
	}
	
	if (!file_exists (buff))
	{
		output_appendf (&debugger.output, "error: file %s does not exist", buff);
		return;
	}

	strcpy (debugger.filename, buff);
	debugger_cmd_reload (args);
}


void debugger_print_msr (void)
{
	__u32 msr_bitmasks[] =
	{
		MSR_POW, MSR_ILE, MSR_EE, MSR_PR, MSR_FP, MSR_ME, MSR_FE0, MSR_SE, MSR_BE,
		MSR_FE1, MSR_IP, MSR_IR, MSR_DR, MSR_RI, MSR_LE,
	};
	char *msr_bitnames[] = 
	{
		"POW", "ILE", "EE", "PR", "FP", "ME", "FE0", "SE", "BE",
		"FE1", "IP", "IR", "DR", "RI", "LE",
	};
	char buff[LINE_WIDTH + 1] = {0};
	int i;
	
	
	for (i = 0; i < sizeof (msr_bitmasks) / 4; i++)
		if (MSR & msr_bitmasks[i])
		{
			strcat (buff, msr_bitnames[i]);
			strcat (buff, " ");
		}

	if (strlen (buff))
		output_appendf (&debugger.output, "MSR = 0x%.8x -> %s", MSR, buff);
	else
		output_appendf (&debugger.output, "MSR = 0x%.8x", MSR);
}


// the result might not fit in one line (only in impossible cases)
void debugger_print_cr (void)
{
	__u32 bitmasks[] =
	{
		CR_LT, CR_GT, CR_EQ, CR_SO,
	};
	char *bitnames[] = 
	{
		"L", "G", "E", "O",
	};
	char buff[2*LINE_WIDTH] = {0};
	char temp1[LINE_WIDTH + 1] = {0};
	char temp2[16];
	int i, j, k;


	for (i = 0; i < 8; i++)
	{
		k = GET_FIELD (CR, i);
		*temp2 = '\0';

		for (j = 0; j < 4; j++)
		{
			if (k & bitmasks[j])
			{
				strcat (temp2, bitnames[j]);
				strcat (temp2, " ");
			}
		}

		if (strlen (temp2))
		{
			temp2[strlen (temp2) - 1] = '\0';
			sprintf (temp1, "%d(%s) ", i, temp2);
			strcat (buff, temp1);
		}
	}

	if (strlen (buff))
		output_appendf (&debugger.output, "CR = 0x%.8x -> %s", CR, buff);
	else
		output_appendf (&debugger.output, "CR = 0x%.8x", CR);
}


void debugger_print_intmask (__u32 b, int index)
{
	char buff[LINE_WIDTH + 1] = {0};


	sprint_intmask (buff, b);
	if (strlen (buff))
		output_appendf (&debugger.output, "%s = 0x%.8x -> %s", variables[index].s, b, buff);
	else
		output_appendf (&debugger.output, "%s = 0x%.8x", variables[index].s, b);
}


void debugger_cmd_reg (char *args)
{
	unsigned int n, val, index, ri;


	if ((n = args_extract_reg (args, &index)))
	{
		ri = reg_indices[index];
		if ((n = args_parse_expression_uint (args + n, &val)))
			REGS (ri) = val;

		switch (ri)
		{
			case I_MSR:
				debugger_print_msr ();
				break;

			case I_CR:
				debugger_print_cr ();
				break;

			case I_PC:
				if (n)
					debugger.code_address = PC;
				else
					output_appendf (&debugger.output, "%s = 0x%.8x", reg_names[index], REGS (ri));
				break;

			default:
				output_appendf (&debugger.output, "%s = 0x%.8x", reg_names[index], REGS (ri));
		}
	}
	else if ((n = args_extract_variable (args, &ri)))
	{
		if ((n = args_parse_expression_uint (args + n, &val)))
			*variables[ri].uip = val;

		switch (ri)
		{
			case VAR_INTSR:
				debugger_print_intmask (INTSR, ri);
				break;
				
			case VAR_INTMR:
				debugger_print_intmask (INTMR, ri);
				break;
		
			default:
				output_appendf (&debugger.output, "%s = 0x%.8x", variables[ri].s, *variables[ri].uip);
		}
	}
	else
		output_append (&debugger.output, "* usage: r reg [value]");
}


void debugger_cmd_exception (char *args)
{
	__u32 d;


	if (args_parse_expression_uint (args, &d))
	{
		cpu_exception (d + 4);
		debugger.code_address = PC;
	}
	else
		output_append (&debugger.output, "* usage: exception cpu_exception");
}


void debugger_cmd_interrupt (char *args)
{
	__u32 d;


	if (args_parse_expression_uint (args, &d))
	{
		pi_interrupt (d + 4);
		debugger.code_address = PC;
	}
	else
		output_append (&debugger.output, "* usage: interrupt interrupt_mask");
}


void sprint_binary (char *buff, unsigned int a)
{
	int i, j;


	buff[0] = '\0';
	for (j = 0; j < 8; j++)
	{
		for (i = 0; i < 4; i++, a <<= 1)
			strcat (buff, (a & 0x80000000) ? "1" : "0");

		if (j < 7)
			strcat (buff, "-");
	}
}


void debugger_cmd_print (char *args)
{
	unsigned int n, k = 0;
	char buff[LINE_WIDTH + 1];


	if ((n = args_parse_expression_uint (args, &k)))
	{
		sprint_binary (buff, k);
		output_appendf (&debugger.output, "0x%.8x\t0b%s\t%u", k, buff, k);
	}
	else
		output_append (&debugger.output, "* usage: print expression");
}


void debugger_cmd_set (char *args)
{
	__u32 n, k, l;


	if ((n = args_extract_variable (args, &k)))
	{
		if ((n = args_extract_uint (args + n, &l)))
		{
			*variables[k].uip = l;

			debugger_cmd_print (args);
			return;
		}
	}

	output_append (&debugger.output, "* usage: set variable");
}


void debugger_cmd_clear (char *args)
{
	output_destroy (&debugger.output);
	debugger.output.pos = 0;
	debugger.output.nfirst_item = 0;

	window_clear (debugger.output_win);
	screen_redraw_window (&debugger.screen, debugger.output_win);
}


void debugger_add_breakpoint (__u32 address)
{
	int i;
	

	for (i = 0; i < MAX_BREAKPOINTS; i++)
		if (!debugger.breakpoints[i])
			break;

	if (i < MAX_BREAKPOINTS)
	{
		debugger.breakpoints[i] = address;
		debugger.breakpoint_enabled[i] = TRUE;
		output_appendf (&debugger.output, "code breakpoint %d set at 0x%.8x", i, address);
	}
	else
		output_append (&debugger.output, "can't set more code breakpoints");
}


// add code breakpoint
void debugger_cmd_bp (char *args)
{
	__u32 b;


	if (!args_parse_expression_uint (args, &b))
		return output_append (&debugger.output, "* usage: bp code_address");

	debugger_add_breakpoint (b);
}


// clear breakpoint
void debugger_cmd_bc (char *args)
{
	unsigned int b;


	if (!args_parse_expression_uint (args, &b))
		return output_append (&debugger.output, "* usage: bc breakpoint_number");

	if (b < MAX_BREAKPOINTS && debugger.breakpoints[b])
	{
		debugger.breakpoints[b] = 0;
		debugger.breakpoint_enabled[b] = FALSE;
		output_appendf (&debugger.output, "breakpoint %d cleared", b);
	}
}


// disable breakpoint
void debugger_cmd_bd (char *args)
{
	unsigned int b;


	if (!args_parse_expression_uint (args, &b))
		return output_append (&debugger.output, "* usage: bd breakpoint_number");

	if (b < MAX_BREAKPOINTS && debugger.breakpoints[b] && debugger.breakpoint_enabled[b])
	{
		debugger.breakpoint_enabled[b] = FALSE;
		output_appendf (&debugger.output, "breakpoint %d disabled", b);
	}
}


// enable breakpoint
void debugger_cmd_be (char *args)
{
	unsigned int b;


	if (!args_parse_expression_uint (args, &b))
		return output_append (&debugger.output, "* usage: be breakpoint_number");

	if (b < MAX_BREAKPOINTS && debugger.breakpoints[b] && !debugger.breakpoint_enabled[b])
	{
		debugger.breakpoint_enabled[b] = TRUE;
		output_appendf (&debugger.output, "breakpoint %d enabled", b);
	}
}


// list all breakpoints
void debugger_cmd_bl (char *args)
{
	int i;


	for (i = 0; i < MAX_BREAKPOINTS; i++)
	{
		if (debugger.breakpoints[i])
		{
			output_appendf (&debugger.output, "code breakpoint %d at 0x%.8x%s",
											i, debugger.breakpoints[i],
											(debugger.breakpoint_enabled[i]) ? "" : " (disabled)");
		}
	}
}


void debugger_cmd_codedump (char *args)
{
	unsigned int n, k;
	char filename[LINE_WIDTH + 1];
	__u32 start = 0, end = 0, t;
	char operand[64], opcode[64];
	unsigned int op;
	FILE *f;

	
	
	if ((n = args_extract_str (args, filename)))
	{
		if ((k = args_parse_expression_uint (args + n, &start)))
		{
			n += k;
			if ((k = args_parse_expression_uint (args + n, &end)))
			{
				if (start > end)
				{
					t = start;
					start = end;
					end = start;
				}
				
				f = fopen (filename, "w+");

				while (start <= end)
				{
					op = BSWAP32 (MEM32_SAFE (start));
					disassemble (opcode, operand, op, start, &t);

					fprintf (f, "%.8x  %.8x  %s    \t%s\n", start, op, opcode, operand);
					start += 4;
				}

				fclose (f);
				output_appendf (&debugger.output, "code dumped to file %s", filename);
				return;
			}
		}
	}
	
	output_append (&debugger.output, "* usage: codedump filename start end");
}


void debugger_cmd_bpex (char *args)
{
	debugger.break_on_exception = !debugger.break_on_exception;
	output_appendf (&debugger.output, "break on exception %s",
									(debugger.break_on_exception) ? "on" : "off");
}


void debugger_cmd_bphw (char *args)
{
	unsigned int n, k;
	char str[LINE_WIDTH + 1];

	
	
	if ((n = args_extract_str (args, str)))
	{
		if (!args_parse_expression_uint (args + n, &k))
			k = -1;

		if (0 == strcasecmp (str, "read"))
		{
			if (k != -1)
				debugger.break_on_uhw_read = k;
			else
				debugger.break_on_uhw_read = !debugger.break_on_uhw_read;
			output_appendf (&debugger.output, "break on unhandled hw read: %s",
											(debugger.break_on_uhw_read) ? "on" : "off");
		}
		else if (0 == strcasecmp (str, "write"))
		{
			if (k != -1)
				debugger.break_on_uhw_write = k;
			else
				debugger.break_on_uhw_write = !debugger.break_on_uhw_write;
			output_appendf (&debugger.output, "break on unhandled hw write: %s",
											(debugger.break_on_uhw_write) ? "on" : "off");
		}
	}
	else
		output_append (&debugger.output, "* usage: bphw read|write (on|off)");
}


void debugger_cmd_xfb (char *args)
{
	__u32 fb;


	if (args_parse_expression_uint (args, &fb))
	{
		video_set_framebuffer (MEM_ADDRESS (fb));
		output_appendf (&debugger.output, "xfb set to 0x%.8x", fb);
	}
	else
		output_append (&debugger.output, "* usage: xfb address");
}


void debugger_cmd_outdump (char *args)
{
	char filename[LINE_WIDTH + 1];
	int i;
	FILE *f;


	if (!args_extract_str (args, filename))
		strcpy (filename, "output.log");

	f = fopen (filename, "w+");
	for (i = 0; i < debugger.output.nitems; i++)
		fprintf (f, "%s\n", debugger.output.items[output_item_index (&debugger.output, i)]);

	fclose (f);
	output_appendf (&debugger.output, "output dumped to file %s", filename);
}


// load binary file to memory
void debugger_cmd_dload (char *args)
{
	char filename[LINE_WIDTH + 1];
	__u32 address, size;
	int n;
	FILE *f;


	if ((n = args_extract_str (args, filename)))
	{
		if (!args_parse_expression_uint (args + n, &address))
			address = debugger.mem_address;

		size = MEM_SIZE - (address & MEM_MASK);

		f = fopen (filename, "rb");
		if (f)
		{
			// loads at most 'size' bytes
			n = fread (MEM_ADDRESS (address), 1, size, f);
			fclose (f);
		
			output_appendf (&debugger.output, "loaded %d bytes from file %s at %.8x",
											n, filename, address);
		}
		else
			output_appendf (&debugger.output, "error: couldn't open file %s", filename);
	}
	else
		output_append (&debugger.output, "* usage: dload filename [address]");
}


void debugger_cmd_map (char *args)
{
	MapItem *mi;


	mi = debugger.map;
	if (!mi)
		output_append (&debugger.output, "map doesn't contain any entries");
	else while (mi)
	{
		output_appendf (&debugger.output, "%x\t%s", mi->address, mi->s);
		mi = mi->next;
	}
}


void debugger_cmd_save (char *args)
{
	char filename[LINE_WIDTH + 1];
	int n, compress;


	if ((n = args_extract_str (args, filename)))
	{
		if (!args_parse_expression_uint (args + n, &compress))
			compress = debugger.compress_states;
	}
	else
	{
		strcpy (filename, debugger.filename);
		kill_extension (filename);
		strcat (filename, ".gcs");
		
		compress = debugger.compress_states;
	}
	
	save_state (filename, compress);
}


void debugger_cmd_dumpstream (char *args)
{
	char filename[LINE_WIDTH + 1];
	__u32 address, length;
	int n, k;


	if ((n = args_extract_str (args, filename)))
	{
		if ((k = args_parse_expression_uint (args + n, &address)))
		{
			n += k;
			if ((k = args_parse_expression_uint (args + n, &length)))
			{
				audio_dump_stream (filename, MEM_ADDRESS (address), length);
				
				output_appendf (&debugger.output, "stream dumped to %s", filename);
				return;
			}
		}
	}

	output_appendf (&debugger.output, "* usage: dumpstream filename address length", filename);
}


void debugger_cmd_dumptexture (char *args)
{
	char filename[LINE_WIDTH + 1];
	__u32 address, format, width, height;
	int n, k;


	if ((n = args_extract_str (args, filename)))
	{
		if ((k = args_parse_expression_uint (args + n, &address)))
		{
			n += k;
			if ((k = args_parse_expression_uint (args + n, &width)))
			{
				n += k;
				if ((k = args_parse_expression_uint (args + n, &height)))
				{
					n += k;
					if ((k = args_parse_expression_uint (args + n, &format)))
					{
						video_dump_texture (filename, MEM_ADDRESS (address), width, height, format);
				
						output_appendf (&debugger.output, "texture dumped to %s", filename);
						return;
					}
				}
			}
		}
	}

	output_appendf (&debugger.output, "* usage: dumptexture filename address width height format", filename);
}


Command commands[] =
{
	{ "help", "list of commands", debugger_cmd_help },
	{ "x", "detach from debugger", debugger_cmd_detach },
	{ "quit", "quit", debugger_cmd_quit },
	{ "n", "step into", debugger_cmd_step_into },
	{ "s", "step over", debugger_cmd_step_over },
	{ "mem", "show memory at given address and write data", debugger_cmd_mem },
	{ "code", "disassemble code at given address", debugger_cmd_code },
	{ "r", "show / change register", debugger_cmd_reg },
	{ "run", "run program", debugger_cmd_run },
	{ "load", "load executable", debugger_cmd_load },
	{ "save", "save state", debugger_cmd_save },
	{ "reload", "reload executable", debugger_cmd_reload },
	{ "print", "print the value of an expression", debugger_cmd_print },
	{ "set", "set variable", debugger_cmd_set },
	{ "exception", "initiate cpu exception", debugger_cmd_exception },
	{ "interrupt", "initiate cpu interrupt", debugger_cmd_interrupt },
	{ "clear", "clear output window", debugger_cmd_clear },
	{ "bp", "set code breakpoint", debugger_cmd_bp },
	{ "bc", "clear breakpoint", debugger_cmd_bc },
	{ "bd", "disable breakpoint", debugger_cmd_bd },
	{ "be", "enable breakpoint", debugger_cmd_be },
	{ "bl", "list breakpoints", debugger_cmd_bl },
	{ "codedump", "dump code to file", debugger_cmd_codedump },
	{ "outdump", "dump output to file", debugger_cmd_outdump },
	{ "bphw", "break on unhandled hardware access", debugger_cmd_bphw },
	{ "dload", "load data file at memory address", debugger_cmd_dload },
	{ "map", "show memory map entries", debugger_cmd_map },
	{ "dumpstream", "dump audio stream to file", debugger_cmd_dumpstream },
	{ "dumptexture", "dump texture to file", debugger_cmd_dumptexture },
	{ "xfb", "set framebuffer address", debugger_cmd_xfb },
	{ "bpex", "break on exception", debugger_cmd_bpex },
};




void gdebug_event (int event, const char *format, ...)
{
	va_list ap;
	
	va_start (ap, format);
	switch (event)
	{
		case EVENT_QUIT:
			gcube_running = FALSE;
			gdebug_running = FALSE;
			emu_state = EMU_EXIT | GDEBUG_EXIT;
			output_vappend (&debugger.output, ap, format);
			break;
		
		case EVENT_EFATAL:
		case EVENT_STOP:
			if (gcube_running)
			{
				gcube_running = FALSE;
				emu_state = GDEBUG_RUN;
			}
			else if (gdebug_running)
			{
				gdebug_running = FALSE;
			}
			output_vappend (&debugger.output, ap, format);
			break;
		
		case EVENT_EMAJOR:
			gdebug_running = FALSE;
			output_vappend (&debugger.output, ap, format);
			break;
		
		case EVENT_INFO:
		case EVENT_LOG:
		case EVENT_REPORT:
		case EVENT_EMINOR:
			output_vappend (&debugger.output, ap, format);
			break;

		case EVENT_EXCEPTION:
			if (debugger.break_on_exception)
			{
				output_vappend (&debugger.output, ap, format);
				gdebug_running = FALSE;
			}
			break;

		case EVENT_UHW_READ:
			output_vappend (&debugger.output, ap, format);
			if (debugger.break_on_uhw_read)
				gdebug_running = FALSE;
			break;

		case EVENT_UHW_WRITE:
			output_vappend (&debugger.output, ap, format);
			if (debugger.break_on_uhw_write)
				gdebug_running = FALSE;
			break;

		case EVENT_TRAP:
			output_vappend (&debugger.output, ap, format);
			gdebug_running = FALSE;
			break;
	}

	va_end (ap);
}


// 'control + c' handler
void sig_ctrl_c (int data)
{
	if (gcube_running)
	{
		gcube_running = FALSE;
		emu_state = GDEBUG_RUN;
	}
	else if (gdebug_running)
	{
		gdebug_running = FALSE;
	}
}


// debugger functions

void debugger_cmd_step_into (char *args)
{
	debugger.last_pc = PC;
	cpu_execute ();
	debugger.code_address = PC;
}


void debugger_cmd_detach (char *args)
{
	emu_state = GDEBUG_EXIT | GCUBE_RUN;
}


void debugger_cmd_quit (char *args)
{
	emu_state = EMU_EXIT | GDEBUG_EXIT;
}


int command_find (char *cmd)
{
	int i;


	for (i = 0; i < (sizeof (commands) / sizeof (Command)); i++)
		if (0 == strncmp (cmd, commands[i].name, strlen (cmd)))
			break;
	
	if (i == (sizeof (commands) / sizeof (Command)))
		return -1;
	else
		return i;
}


void debugger_cmd_help (char *args)
{
	int i;
	

	output_appendf (&debugger.output, "");
	output_appendf (&debugger.output, "available commands:");
	for (i = 0; i < (sizeof (commands) / sizeof (Command)); i++)
		output_appendf (&debugger.output, "  %s  \t\t%s", commands[i].name, commands[i].description);
}


void debugger_parse_command (char *command)
{
	char cmd[LINE_WIDTH + 1] = {0}, args[LINE_WIDTH + 1] = {0};
	int i = 0;


	sscanf (command, "%s %[^\n]", cmd, args);

	i = command_find (cmd);
	if (i < 0)
		output_appendf (&debugger.output, "unknown command: %s", cmd);
	else
		commands[i].action (args);

	screen_redraw (&debugger.screen);
}


void mem_show (Window *win, int lines, unsigned int addr)
{
	int i, j, c;


	for (i = win->nlines - lines; i < win->nlines; i++)
	{
		window_cprintf (win, 0, i, FG_RED, "%.8x", addr);
		window_line_set_attribs (win, i, 10, 60, FG_GREEN, 0);
		window_line_set_attribs (win, i, 60, 80, FG_BLACK, 0);
		for (j = 0; j < 0x10; j++)
		{
			c = MEM_SAFE (addr + j);
			window_printf (win, 10 + j*3 + j/4, i, "%.2x", c);
			window_printf (win, 63 + j, i, "%c", (c >= 0x20 && c < 0x7f) ? c : '.');
		}
		
		addr += 0x10;
	}
}


MapItem *debugger_map_find_item (__u32 address)
{
	MapItem *mi;


	mi = debugger.map;
	while (mi)
	{
		if (mi->address == address)
			break;
		
		mi = mi->next;
	}
	
	return mi;
}


void code_show (Window *win, int lines, unsigned int addr)
{
	__u32 target;
	char operand[64], opcode[64];
	unsigned int op;
	int i, k;
	MapItem *mi;


	addr -= (win->nlines/2) * 4;
	for (i = win->nlines - lines; i < win->nlines; i++)
	{
		op = BSWAP32 (MEM32_SAFE (addr));
		disassemble (opcode, operand, op, addr, &target);
		window_clear_line (win, i);
		
		for (k = 0; k < MAX_BREAKPOINTS; k++)
			if (debugger.breakpoint_enabled[k] && addr == debugger.breakpoints[k])
				break;

		if (k < MAX_BREAKPOINTS)
			window_line_set_attribs (win, i, 0, 80, FG_YELLOW, ATTRIB_STANDOUT);
		else if (addr == PC)
			window_line_set_attribs (win, i, 0, 80, FG_MAGENTA, ATTRIB_STANDOUT);
		else
		{
			window_line_set_attribs (win, i, 0, 8, FG_RED, 0);
			window_line_set_attribs (win, i, 10, 18, FG_GREEN, 0);
			window_line_set_attribs (win, i, 20, 60, FG_BLUE, 0);
			window_line_set_attribs (win, i, 60, 80, FG_BLACK, ATTRIB_BOLD);
		}
		window_printf (win, 0, i, "%.8x  %.8x  %s", addr, op, opcode);

		if ((target != addr + 4) && (mi = debugger_map_find_item (target)))
			window_printf (win, 32, i, "%s", mi->s);
		else
			window_printf (win, 32, i, "%s", operand);

		if ((mi = debugger_map_find_item (addr)))
			window_printf (win, 60, i, "%s", mi->s);
	
		addr += 4;
	}

	if (debugger.code_cmd_input)
	{
		window_title_clear (win, 7, LINE_WIDTH);
		window_title_printf (win, CODE_CMD_CURSOR_OFFSET, debugger.code_cmd);
	}
}


void regs_show (Window *win, int sw)
{
	static __u32 oldregs[4096] = {0};
	static __u32 row1[] = { I_CR, I_FPSCR, I_XER, I_MSR, I_CTR, I_HID0, I_HID1, I_HID2 };
	static int fgc[] = { FG_BLACK, FG_MAGENTA };
	int i, j;


	window_cprintf (win, 0, 0, FG_BLUE, "gpr");
	for (j = 0; j < 4; j++)
	{
		for (i = 0; i < 8; i++)
		{
			window_cprintf (win, 8 + i*9, j, fgc[oldregs[j*8 + i] != GPR[j*8 + i]],
											"%.8x", GPR[j*8 + i]);
			oldregs[j*8 + i] = GPR[j*8 + i];
		}
	}

	window_cprintf (win, 8, 4, FG_BLUE, "cr       fpscr    xer      msr      ctr      hid0     hid1     hid2");
	window_printf (win, 15, 4, (GET_CR0 & CR_EQ) ? "=" : (GET_CR0 & CR_LT) ? "<" : (GET_CR0 & CR_GT) ? ">" : "");
	for (i = 0; i < 8; i++)
	{
		window_cprintf (win, 8 + i*9, 5, fgc[oldregs[row1[i]] != REGS (row1[i])],
										"%.8x", REGS (row1[i]));
		oldregs[row1[i]] = REGS (row1[i]);
	}

	window_cprintf (win, 8, 6, FG_BLUE, "dec      ic                                           lr       pc");
	window_printf (win, 8, 7, "%.8x %.8x                                              %.8x", DEC, IC, PC);
	window_cprintf (win, 62, 7, fgc[oldregs[I_LR] != REGS (I_LR)],
									"%.8x", REGS (I_LR));
	oldregs[I_LR] = REGS (I_LR);
}


void output_show (Window *win, OutputWin *out, int nlines, int pos)
{
	int i;
	float t;


	for (i = win->nlines - nlines; i < win->nlines; i++)
	{
		if (pos >= out->nitems)
			break;

		window_clear_line (win, i);
		window_printf (win, 3, i, out->items[output_item_index (out, pos++)]);
	}
	
	t = (float) 100.0 * (out->pos + win->nlines) / out->nitems;
	if (t > 100.0)
		t = 100.0;

	window_title_printf (win, LINE_WIDTH - 7, " %3.0f %% ", t);
}


void output_append (OutputWin *out, char *str)
{
	if (out->nitems >= MAX_OUTPUT_LINES)
	{
		free (out->items[output_item_index (out, --out->nitems)]);
		out->nfirst_item++;
		if (out->nfirst_item >= MAX_OUTPUT_LINES)
			out->nfirst_item = 0;
	}

	out->items[output_item_index (out, out->nitems++)] = strdup (str);

	if (out->visitems)
		out->pos = out->nitems - out->visitems;
	if (out->pos < 0)
		out->pos = 0;
}


void output_vappend (OutputWin *out, va_list ap, const char *format)
{
	char buff[LINE_WIDTH + 1];
	
	
	vsnprintf (buff, sizeof (buff), format, ap);
	output_appendf (out, buff);
}


void output_appendf (OutputWin *out, char *format, ...)
{
	char str[LINE_WIDTH + 1];
	va_list ap;


	va_start (ap, format);
	vsnprintf (str, sizeof (str), format, ap);
	va_end (ap);

	output_append (out, str);
}


void output_destroy (OutputWin *out)
{
	while (out->nitems)
		free (out->items[--out->nitems]);
}


void command_history_append (CommandWin *cmd, char *str)
{
	if (cmd->hlines < MAX_HISTORY_LINES)
		if (!cmd->hlines || (cmd->hlines && strcmp (cmd->history[cmd->hlines-1], str)))
			cmd->history[cmd->hlines++] = strdup (str);
}


void command_history_destroy (CommandWin *cmd)
{
	while (cmd->hlines)
		free (cmd->history[--cmd->hlines]);
}


void debugger_execute_command (void)
{
	if (debugger.console.hpos)
	{
		// execute history command
		debugger_parse_command (debugger.console.history[debugger.console.hlines - debugger.console.hpos]);
		debugger.console.hpos = 0;
	}
	else if (strlen (debugger.console.buff) > 0)
	{
		// execute buffer command
		command_history_append (&debugger.console, debugger.console.buff);
		debugger_parse_command (debugger.console.buff);
	}
	else if (debugger.console.hlines)
		// execute last command
		debugger_parse_command (debugger.console.history[debugger.console.hlines - 1]);

	memset (debugger.console.buff, 0, CONSOLE_MAX_CHARS);
	debugger.console.pos = 0;
	window_refresh (debugger.console_win);
}



// window handling functions

void on_mem_resize (Window *win, int nlines, void *data)
{
	mem_show (win, win->nlines, debugger.mem_address);
	
	debugger.mem_lines = nlines;
}


void on_mem_refresh (Window *win, void *data)
{
	mem_show (win, win->nlines, debugger.mem_address);
}


void on_mem_scroll (Window *win, int nlines, int direction, void *data)
{
	unsigned int addr;
	int i;


	if (direction == DIR_DOWN)
	{
		addr = debugger.mem_address + win->nlines * 0x10;
		debugger.mem_address += 0x10 * nlines;
	}
	else
	{
		addr = debugger.mem_address - 0x10;
		debugger.mem_address -= 0x10 * nlines;
	}

	for (i = 0; i < nlines; i++)
	{
		if (direction == DIR_DOWN)
		{
			window_roll (win, 1, direction);
			mem_show (win, 1, addr);
			addr += 0x10;
		}
		else
		{
			mem_show (win, 1, addr);
			window_roll (win, 1, direction);
			addr -= 0x10;
		}
	}
}


int buff_keystroke (char *buff, int *pos, int maxh, int key)
{
	int i;


	if (key >= 0x20 && key < 0x7f)
	{
		if (*pos < maxh)
			buff[(*pos)++] = key;

		return TRUE;
	}
	else switch (key)
	{
		case KEY_BACKSPACE:
			if (*pos > 0)
			{
				for (i = --(*pos); i < strlen (buff); i++)
					buff[i] = buff[i + 1];
				buff[i] = '\0';
			}
			return TRUE;
			
		case KEY_DC:
				for (i = *pos; i < strlen (buff); i++)
					buff[i] = buff[i + 1];
				buff[i] = '\0';
			return TRUE;

		case KEY_LEFT:
			if (*pos > 0)
				(*pos)--;
			return TRUE;

		case KEY_RIGHT:
			if (*pos < strlen (buff))
				(*pos)++;
			return TRUE;

		case KEY_HOME:
			*pos = 0;
			return TRUE;

		case KEY_END:
			*pos = strlen (buff);
			return TRUE;

		default:
			return FALSE;
	}
}


int on_code_keypress (Window *win, int key, int modifiers, void *data)
{
	if (!debugger.code_cmd_input)
	{
		switch (key)
		{
			case ':':
				// label
				debugger.code_cmd_input = CODE_CMD_SET_LABEL;
				screen_set_cursor_relative (&debugger.screen, win, debugger.hcodepos + CODE_CMD_CURSOR_OFFSET, 0);
				return TRUE;
			
			case 'b':
				{
					int k;

					for (k = 0; k < MAX_BREAKPOINTS; k++)
						if (debugger.breakpoint_enabled[k] && debugger.code_address == debugger.breakpoints[k])
							break;

					if (k < MAX_BREAKPOINTS)
					{
						debugger.breakpoint_enabled[k] = FALSE;
						debugger.breakpoints[k] = 0;
						
						output_appendf (&debugger.output, "code breakpoint %d cleared", k);
					}
					else
						debugger_add_breakpoint (debugger.code_address);

					screen_redraw (&debugger.screen);
				}
				return TRUE;
			
			default:
				return FALSE;
		}
	}
	else switch (key)
	{
		// enter
		case 0x0d:
			switch (debugger.code_cmd_input)
			{
				case CODE_CMD_SET_LABEL:
					debugger_map_add_item (debugger.code_address, debugger.code_cmd);
					debugger.map_changed = TRUE;
					break;
			}
			
			memset (debugger.code_cmd, 0, sizeof (debugger.code_cmd));
			debugger.code_cmd_input = 0;
			debugger.hcodepos = 0;

			window_title_clear (win, 7, LINE_WIDTH);
			screen_redraw_window (&debugger.screen, win);
			screen_set_cursor_relative (&debugger.screen, win, 0, 0);
			return TRUE;

		default:
			if (buff_keystroke (debugger.code_cmd, &debugger.hcodepos, CODE_CMD_MAX_CHARS, key))
			{
				screen_redraw_window (&debugger.screen, win);
				screen_set_cursor_relative (&debugger.screen, win, debugger.hcodepos + CODE_CMD_CURSOR_OFFSET, 0);
				return TRUE;
			}
			else
				return FALSE;
	}

	return FALSE;
}


void on_code_resize (Window *win, int nlines, void *data)
{
	code_show (win, win->nlines, debugger.code_address);
	
	debugger.code_lines = nlines;
}


void on_code_refresh (Window *win, void *data)
{
	code_show (win, win->nlines, debugger.code_address);
}


void on_code_scroll (Window *win, int nlines, int direction, void *data)
{
	unsigned int addr;
	int i;


	if (direction == DIR_DOWN)
	{
		addr = debugger.code_address + win->nlines * 4;
		debugger.code_address += 4 * nlines;
	}
	else
	{
		addr = debugger.code_address - 4;
		debugger.code_address -= 4 * nlines;
	}

	for (i = 0; i < nlines; i++)
	{
		if (direction == DIR_DOWN)
		{
			window_roll (win, 1, direction);
			code_show (win, 1, addr);
			addr += 4;
		}
		else
		{
			code_show (win, 1, addr);
			window_roll (win, 1, direction);
			addr -= 4;
		}
	}
}


void on_regs_refresh (Window *win, void *data)
{
	regs_show (win, (int) data);
}


int on_keypress (Screen *win, int key, int modifiers, void *data)
{
	switch (key)
	{
		case KEY_F(5):
			debugger_cmd_run (NULL);
			screen_redraw (&debugger.screen);
			return TRUE;

		case KEY_F(8):
		case 'n':
			debugger_cmd_step_into (NULL);
			screen_redraw (&debugger.screen);
			return TRUE;
		
		case KEY_F(9):
		case 's':
			debugger_cmd_step_over (NULL);
			screen_redraw (&debugger.screen);
			return TRUE;
		
		case KEY_F(11):
		case 'x':
			debugger_cmd_detach (NULL);
			screen_redraw (&debugger.screen);
			return TRUE;

		case KEY_F(12):
		case 'q':
			debugger_cmd_quit (NULL);
			return TRUE;
		
		default:
			return FALSE;
	}
}


int on_unhandled_keypress (Screen *win, int key, int modifiers, void *data)
{
	screen_select_window (&debugger.screen, SELECT_N0 + 4);
	return window_handle_keypress (debugger.console_win, key, modifiers);
}


void on_output_resize (Window *win, int nlines, void *data)
{
	if (nlines > debugger.output.nitems - debugger.output.pos)
	{
		debugger.output.pos -= nlines - (debugger.output.nitems - debugger.output.pos);

		if (debugger.output.pos < 0)
			debugger.output.pos = 0;
	}

	output_show (win, &debugger.output, nlines, debugger.output.pos);
	debugger.output.visitems = win->nlines;
	
	debugger.output_lines = nlines;
}


void on_output_refresh (Window *win, void *data)
{
	output_show (win, &debugger.output, win->nlines, debugger.output.pos);
}


void on_output_scroll (Window *win, int nlines, int direction, void *data)
{
	int pos, i;


	if (!debugger.output.nitems || debugger.output.nitems < win->nlines)
		return;

	if (direction == DIR_FIRST)
	{
		debugger.output.pos = 0;
		output_show (win, &debugger.output, win->nlines, debugger.output.pos);
	}
	else if (direction == DIR_LAST)
	{
		debugger.output.pos = debugger.output.nitems - win->nlines;
		output_show (win, &debugger.output, win->nlines, debugger.output.pos);
	}
	else if (direction == DIR_DOWN)
	{
		pos = debugger.output.pos + win->nlines;
		if (nlines >= debugger.output.nitems - pos)
			nlines = debugger.output.nitems - pos;

		debugger.output.pos += nlines;

		for (i = 0; i < nlines; i++)
		{
			window_roll (win, 1, direction);
			output_show (win, &debugger.output, 1, pos);
			pos++;
		}
	}
	else
	{
		pos = debugger.output.pos - 1;
		if (nlines > debugger.output.pos)
			nlines = debugger.output.pos;

		debugger.output.pos -= nlines;

		for (i = 0; i < nlines; i++)
		{
			output_show (win, &debugger.output, 1, pos);
			window_roll (win, 1, direction);
			pos--;
		}
	}
}


void on_command_refresh (Window *win, void *data)
{
	window_clear_line (win, 0);
	if (debugger.console.hpos)
		window_printf (win, 0, 0, ".command: %s", debugger.console.history[debugger.console.hlines - debugger.console.hpos]);
	else
		window_printf (win, 0, 0, ".command: %s", debugger.console.buff);
}


void on_command_enter (Window *win, void *data)
{
	screen_set_cursor_relative (&debugger.screen, win, debugger.console.pos + CONSOLE_CURSOR_OFFSET, 0);
}


int on_command_keypress (Window *win, int key, int modifiers, void *data)
{
	// if looking at history buffer entry, copy it to edit buffer
	if (debugger.console.hpos &&
			((key >= 0x20 && key < 0x7f) || (key == KEY_BACKSPACE) || (key == KEY_DC)))
	{
		strcpy (debugger.console.buff, debugger.console.history[debugger.console.hlines - debugger.console.hpos]);
		debugger.console.hpos = 0;
	}

	switch (key)
	{
		// enter
		case 0x0d:
			debugger_execute_command ();
			screen_redraw_window (&debugger.screen, win);
			screen_set_cursor_relative (&debugger.screen, win, debugger.console.pos + CONSOLE_CURSOR_OFFSET, 0);
			return TRUE;

		case KEY_DOWN:
			if (debugger.console.hpos > 0)
			{
				debugger.console.hpos--;
				if (debugger.console.hpos > 0)
					debugger.console.pos = strlen (debugger.console.history[debugger.console.hlines - debugger.console.hpos]);
				else
					debugger.console.pos = strlen (debugger.console.buff);
				screen_redraw_window (&debugger.screen, win);
				screen_set_cursor_relative (&debugger.screen, win, debugger.console.pos + CONSOLE_CURSOR_OFFSET, 0);
			}
			return TRUE;
			
		case KEY_UP:
			if (debugger.console.hpos < debugger.console.hlines)
			{
				debugger.console.hpos++;
				debugger.console.pos = strlen (debugger.console.history[debugger.console.hlines - debugger.console.hpos]);

				screen_redraw_window (&debugger.screen, win);
				screen_set_cursor_relative (&debugger.screen, win, debugger.console.pos + CONSOLE_CURSOR_OFFSET, 0);
			}
			return TRUE;

		case KEY_NPAGE:
		case KEY_PPAGE:
			if (!debugger.output_win->hidden && debugger.output_win->nlines > 1)
			{
				window_scroll (debugger.output_win, debugger.output_win->nlines - 1,
											 (key == KEY_NPAGE) ? DIR_DOWN : DIR_UP);
				screen_refresh (&debugger.screen);
				return TRUE;
			}
			else
				return FALSE;

		case KEY_END:
			if (debugger.console.hpos)
				debugger.console.pos = strlen (debugger.console.history[debugger.console.hlines - debugger.console.hpos]);
			else
				debugger.console.pos = strlen (debugger.console.buff);
			screen_redraw_window (&debugger.screen, win);
			screen_set_cursor_relative (&debugger.screen, win, debugger.console.pos + CONSOLE_CURSOR_OFFSET, 0);
			return TRUE;

		case KEY_RIGHT:
			if (debugger.console.hpos)
			{
				if (debugger.console.pos < strlen (debugger.console.history[debugger.console.hlines - debugger.console.hpos]))
				{
					debugger.console.pos++;
					screen_redraw_window (&debugger.screen, win);
					screen_set_cursor_relative (&debugger.screen, win, debugger.console.pos + CONSOLE_CURSOR_OFFSET, 0);
				}
				
				return TRUE;
			}
			// fall through
		default:
			if (buff_keystroke (debugger.console.buff, &debugger.console.pos, CONSOLE_MAX_CHARS, key))
			{
				screen_redraw_window (&debugger.screen, win);
				screen_set_cursor_relative (&debugger.screen, win, debugger.console.pos + CONSOLE_CURSOR_OFFSET, 0);
				return TRUE;
			}
			else
				return FALSE;
	}
}


MapItem *map_item_create (__u32 address, char *s)
{
	MapItem *mi;


	if (!s)
		return NULL;

	mi = calloc (1, sizeof (MapItem));
	mi->address = address;
	mi->s = strdup (s);
	
	return mi;	
}


void map_item_destroy (MapItem *mi)
{
	if (mi)
	{
		free (mi->s);
		free (mi);
	}
}


void debugger_map_add_item (__u32 address, char *s)
{
	MapItem *mi;


	// if an entry with specified address already exists
	// overwrite it (or delete it, if s is empty)
	mi = debugger_map_find_item (address);
	if (mi)
	{
		if (strlen (s))
		{
			free (mi->s);
			mi->s = strdup (s);
		}
		else
			map_item_destroy (mi);
			
		return;
	}

	mi = map_item_create (address, s);
	if (debugger.map)
	{
		debugger.map_last->next = mi;
		mi->prev = debugger.map_last;
		debugger.map_last = mi;
	}
	else
	{
		debugger.map = mi;
		debugger.map_last = mi;
	}
}


void debugger_map_remove_item (MapItem *mi)
{
	if (!mi)
		return;

	if (mi == debugger.map)
	{
		debugger.map = NULL;
		debugger.map_last = NULL;

		map_item_destroy (mi);
	}
	else if (mi == debugger.map_last)
	{
		debugger.map_last = mi->prev;
		if (mi->prev)
			mi->prev->next = NULL;

		map_item_destroy (mi);
	}
	else
	{
		if (mi->prev)
			mi->prev->next = mi->next;
		if (mi->next)
			mi->next->prev = mi->prev;

		map_item_destroy (mi);
	}
}


void debugger_map_destroy (void)
{
	while (debugger.map_last)
		debugger_map_remove_item (debugger.map_last);
}

//
void sprint_intmask (char *buff, __u32 b)
{
	__u32 bitmasks[] =
	{
		INTERRUPT_HSP, INTERRUPT_DEBUG, INTERRUPT_CP, INTERRUPT_PE_FINISH,
		INTERRUPT_PE_TOKEN, INTERRUPT_VI, INTERRUPT_MEM, INTERRUPT_DSP,
		INTERRUPT_AI, INTERRUPT_EXI, INTERRUPT_SI, INTERRUPT_DI,
		INTERRUPT_RSW, INTERRUPT_ERROR,
	};
	char *bitnames[] = 
	{
		"HSP", "DBG", "CP", "PFIN", "PTOK", "VI", "MEM", "DSP", "AI",
		"EXI", "SI", "DI", "RSW", "ERROR",
	};
	int i;


	for (i = 0; i < sizeof (bitmasks) / 4; i++)
		if (b & bitmasks[i])
		{
			strcat (buff, bitnames[i]);
			strcat (buff, " ");
		}

	i = strlen (buff);
	// kill last space
	if (i)
		buff[i - 1] = '\0';
}


void gdebug_print_intmask (__u32 mask, char *msg)
{
	char buff[256] = {0};


	sprint_intmask (buff, INTSR);
	DEBUG (EVENT_INFO, "%s%s", msg, buff);
}


void gdebug_load_map (char *filename)
{
	char buff[256];
	FILE *map;
	__u32 address;
	int i = 0;


	sprintf (buff, "%s.map", filename);
	if (!file_exists (buff))
	{
		kill_extension (filename);
		sprintf (buff, "%s.map", filename);

		if (!file_exists (buff))
			return;
	}

	strcpy (debugger.map_filename, buff);	

	map = fopen (buff, "r");
	
	while (!feof (map))
	{
		if (2 == fscanf (map, "%x %s\n", &address, buff))
		{
			debugger_map_add_item (address, buff);
			i++;
		}
		else
		// ignore line
			fscanf (map, "%*[^\n]\n");
	}
	
	fclose (map);

	output_appendf (&debugger.output, "loaded map file %s with %d entries", debugger.map_filename, i);
}


void gdebug_save_map (void)
{
	FILE *f;
	MapItem *mi;


	if (!debugger.map || !debugger.map_changed)
		return;

	if (!*debugger.map_filename)
	{
		strcpy (debugger.map_filename, debugger.filename);
		kill_extension (debugger.map_filename);
		strcat (debugger.map_filename, ".map");
	}
	
	f = fopen (debugger.map_filename, "w");

	mi = debugger.map;
	while (mi)
	{
		fprintf (f, "%x %s\n", mi->address, mi->s);
		mi = mi->next;
	}

	fclose (f);

	output_appendf (&debugger.output, "saved map to file %s", debugger.map_filename);
}


void gdebug_run (__u32 pc)
{
	Window *win;


	debugger.code_address = pc;

	screen_init (&debugger.screen, debugger.colors);

	if (!debugger.initialized)
	{
		debugger.mem_lines = 3;
		debugger.code_lines = 5;
		debugger.output_lines = 3;

		debugger.refresh_rate = 100000;
		debugger.compress_states = TRUE;

		debugger.initialized = TRUE;
	}

	debugger.regs_win = win = window_create (0, 0, 8, ".regs:");
	win->fixed_size = TRUE;
	if (debugger.regs_hidden)
		win->hidden = TRUE;
	window_add_refresh_callback (win, on_regs_refresh, NULL);
	screen_add_window (&debugger.screen, win);

	debugger.mem_win = win = window_create (0, 0, debugger.mem_lines, ".mem:");
	if (debugger.mem_hidden)
		win->hidden = TRUE;
	window_add_refresh_callback (win, on_mem_refresh, NULL);
	window_add_size_change_callback (win, on_mem_resize, NULL);
	window_add_scroll_callback (win, on_mem_scroll, NULL);
	screen_add_window (&debugger.screen, win);

	debugger.code_win = win = window_create (0, 0, debugger.code_lines, ".code:");
	if (debugger.code_hidden)
		win->hidden = TRUE;
	window_add_keypress_callback (win, on_code_keypress, NULL);
	window_add_refresh_callback (win, on_code_refresh, NULL);
	window_add_size_change_callback (win, on_code_resize, NULL);
	window_add_scroll_callback (win, on_code_scroll, NULL);
	screen_add_window (&debugger.screen, win);

	debugger.output_win = win = window_create (0, 0, debugger.output_lines, ".output:");
	debugger.output.visitems = debugger.output_lines;
	if (debugger.output_hidden)
		win->hidden = TRUE;
	window_add_refresh_callback (win, on_output_refresh, NULL);
	window_add_size_change_callback (win, on_output_resize, NULL);
	window_add_scroll_callback (win, on_output_scroll, NULL);
	screen_add_window (&debugger.screen, win);

	debugger.console_win = win = window_create (0, 0, 1, ".command:");
	win->no_title = TRUE;
	win->fixed_size = TRUE;
	win->always_visible = TRUE;
	window_add_keypress_callback (win, on_command_keypress, NULL);
	window_add_enter_callback (win, on_command_enter, NULL);
	window_add_refresh_callback (win, on_command_refresh, NULL);
	screen_add_window (&debugger.screen, win);

	screen_collapse (&debugger.screen, 0);
	screen_expand (&debugger.screen, 0);

	screen_add_keypress_callback (&debugger.screen, on_keypress, NULL);
	screen_add_unhandled_keypress_callback (&debugger.screen, on_unhandled_keypress, NULL);
	screen_redraw (&debugger.screen);
	screen_select_window (&debugger.screen, SELECT_LAST);

	while (!(emu_state & GDEBUG_EXIT))
		screen_handle_input (&debugger.screen, getch ());

	debugger.regs_hidden = debugger.regs_win->hidden;
	debugger.mem_hidden = debugger.mem_win->hidden;
	debugger.code_hidden = debugger.code_win->hidden;
	debugger.output_hidden = debugger.output_win->hidden;
	screen_cleanup (&debugger.screen);
	
	emu_state &= ~GDEBUG_EXIT;
}
#endif // GDEBUG


// too much of refresh stuff

__u32 refresh_delay = 640*480*3;

void do_refresh (void)
{
	if (IC > refresh_delay)
	{
		IC = 0;
		vi_refresh ();
	}
}


// refresh will be done manually by gcube_force_refresh
// if the call won't be made for a long time
// change to automatic slow refresh
void refresh_manual (void)
{
	if (IC > 640*480*30)
	{
		refresh_delay = 640*480*5;
		vid_refresh = do_refresh;
	}
}


void gcube_refresh_vfast (void)
{
	refresh_delay = 640*480*2;
	vid_refresh = do_refresh;
}


void gcube_refresh_force (void)
{
	IC = 0;
	vi_refresh ();
}


// manual refresh
void gcube_refresh_manual (void)
{
	vid_refresh = refresh_manual;
	// don't call vi interrupt too often
	if (IC > 0xfff)
		gcube_refresh_force ();
}


void gcube_run (void)
{
	gcube_running = TRUE;
	while (gcube_running)
	{
		cpu_execute ();
		
		vid_refresh ();
	}
}


__u32 gcube_load_file (char *filename)
{
	mem_reset ();
	cpu_init ();
	hw_init ();

	PC = load_gc (filename);
	if (PC == 0)
		return 0;

	if (is_gcs (filename))
		vi_reinit_video ();
	else
	// hw_init must be called after the file is loaded
		hw_init ();

	return PC;
}


void gcube_quit (char *msg)
{
#ifdef GDEBUG
	gdebug_running = FALSE;
#endif
	gcube_running = FALSE;
	emu_state = EMU_EXIT | GDEBUG_EXIT;

	quit_message = msg;
}


int main (int argc, char **argv)
{
	if (argc < 2)
	{
#ifdef GDEBUG
		printf ("usage: %s file.dol [color_mode]\n", argv[0]);
#else
		printf ("usage: %s file.dol\n", argv[0]);
#endif
		return 0;
	}

	DEBUG (EVENT_INFO, "%s (c) 2004 by monk", GCUBE_DESCRIPTION);
	if (!gcube_load_file (argv[1]))
	{
		printf ("can't open file %s\n", argv[1]);
		return 0;
	}

	config_load (".gcube");
	vid_refresh = do_refresh;

#ifdef OPEN_WINDOW_ON_START
	video_init (640, 480);
#endif

#ifdef GDEBUG
	debugger.colors = TRANSPARENT_BG;
	if (argc > 2)
		sscanf (argv[2], "%d", &debugger.colors);

	strcpy (debugger.filename, argv[1]);
	gdebug_load_map (argv[1]);

	debugger.starting_pc = PC;
	signal (SIGINT, sig_ctrl_c);

	while (!(emu_state & EMU_EXIT))
	{
		emu_state |= EMU_EXIT;

		if (emu_state & GCUBE_RUN)
			gcube_run ();

		if (emu_state & GDEBUG_RUN)
			gdebug_run (PC);
	}

	gdebug_save_map ();

	output_destroy (&debugger.output);
	command_history_destroy (&debugger.console);
#else
	gcube_run ();
#endif // GDEBUG

	if (quit_message)
		printf ("%s\n", quit_message);

	config_save (".gcube");

	return FALSE;
}
