#ifndef __GCUBE_H
#define __GCUBE_H 1


#define GCUBE_DESCRIPTION		"gcube v0.1"


#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>

#include "general.h"
#include "config.h"

#include "ppc_disasm.h"
#include "cpu.h"
#include "diskio.h"
#include "hw.h"

#include "txlib.h"


#define GDEBUG_RUN				0x001
#define GDEBUG_RUNNING		0x002
#define GDEBUG_EXIT				0x004
#define GCUBE_RUN					0x010
#define GCUBE_RUNNING			0x020
#define GCUBE_EXIT				0x040
#define EMU_EXIT					0x100

#define EVENT_INFO				0
#define EVENT_QUIT				1
#define EVENT_REPORT			2
#define EVENT_TRAP				3
#define EVENT_LOG					4
#define EVENT_EFATAL			8
#define EVENT_EMAJOR			9
#define EVENT_EMINOR			10
#define EVENT_EXCEPTION		100
#define EVENT_UHW_READ		101
#define EVENT_UHW_WRITE		102
#define EVENT_STOP				200

#define CONSOLE_CURSOR_OFFSET			10
#define CONSOLE_MAX_CHARS					(LINE_WIDTH - CONSOLE_CURSOR_OFFSET - 1)
#define CODE_CMD_CURSOR_OFFSET		7
#define CODE_CMD_MAX_CHARS				(LINE_WIDTH - CODE_CMD_CURSOR_OFFSET - 1)

#define MAX_HISTORY_LINES					4096
#define MAX_OUTPUT_LINES					409600
#define MAX_BREAKPOINTS						16
#define MAX_MAP_ITEMS							1024

#define CODE_CMD_SET_LABEL				1

#define VAR_INTSR				0
#define VAR_INTMR				1


typedef struct
{
	char buff[LINE_WIDTH+1];
	int pos;
	char *history[MAX_HISTORY_LINES+1];
	int hlines;
	int hpos;
} CommandWin;


typedef struct
{
	char *items[MAX_OUTPUT_LINES];
	int nitems;
	int visitems;
	int pos;
	
	int nfirst_item;
} OutputWin;


typedef struct MapItemTag
{
	__u32 address;
	char *s;
	struct MapItemTag *prev, *next;
} MapItem;


typedef struct
{
	Screen screen;
	CommandWin console;
	OutputWin output;
	Window *regs_win, *mem_win, *code_win, *output_win, *console_win;

	__u32 mem_address;
	__u32 code_address;

	int initialized;
	unsigned int mem_lines, code_lines, output_lines;
	int regs_hidden, mem_hidden, code_hidden, output_hidden;

	__u32 starting_pc;
	__u32 last_pc;

	int novis;
	__u32 refresh_rate;
	int compress_states;
	
	__u32 breakpoints[MAX_BREAKPOINTS];
	int breakpoint_enabled[MAX_BREAKPOINTS];
	int break_on_exception;
	int break_on_uhw_read;
	int break_on_uhw_write;

	char filename[256];

	MapItem *map, *map_last;
	char map_filename[256];
	int map_changed;

	int code_cmd_input;
	char code_cmd[LINE_WIDTH];
	int hcodepos;
	
	int colors;
} DebuggerState;


typedef struct
{
	char *name;
	char *description;
	void (*action) (char *args);
} Command;


typedef struct
{
	const char *s;
	unsigned int ui;
} UIDef;


typedef struct
{
	const char *s;
	unsigned int *uip;
} UIPDef;


#ifdef GDEBUG
# define DEBUG										gdebug_event
#else
# define DEBUG(X,format,...)			{}
#endif



void gcube_run (void);
void gcube_quit (char *msg);
__u32 gcube_load_file (char *filename);

void gcube_refresh_vfast (void);
void gcube_refresh_manual (void);

#ifdef GDEBUG
void gdebug_event (int event, const char *format, ...);
void gdebug_run (__u32 pc);
void gdebug_print_intmask (__u32 mask, char *msg);
#else
# define gdebug_print_intmask(mask,msg)		{}
#endif


#endif // __GCUBE_H
