#ifndef __HW_SI_H
#define __HW_SI_H 1


#include "general.h"
#include "mem.h"

#define RSI_SIZE				0x100
#define RSI_MASK				0xff

// reversed addressing space

#define RSI8(X)					(*((__u8 *) &rsi[RSI_SIZE - ((X) & RSI_MASK) - 1]))
#define RSI32(X)				(*((__u32 *) &rsi[RSI_SIZE - ((X) & RSI_MASK) - 4]))

extern __u8 rsi[RSI_SIZE];


#define SI_CH0_CMD			0x6400
#define SI_CH0_INL			0x6404
#define SI_CH0_INU			0x6408
#define SI_CH1_CMD			0x640c
#define SI_CH1_INL			0x6410
#define SI_CH1_INU			0x6414
#define SI_CH2_CMD			0x6418
#define SI_CH2_INL			0x641c
#define SI_CH2_INU			0x6420
#define SI_CH3_CMD			0x6424
#define SI_CH3_INL			0x6428
#define SI_CH3_INU			0x642c


#define SI_CHL(B,X,Y)			((B << 16) | ((__u8) (0x80+X) << 8) | ((__u8) (0x80-Y)))
#define SI_CHU(X,Y,L,R)		(((__u8) (0x80+X) << 24) | ((__u8) (0x80+Y) << 16) | (L << 8) | R)

#define SI_POLL									0x6430
#define SI_COMMCSR							0x6434
#define SI_SR										0x6438
#define SI_IO_BUFF							0x6480

#define SI_POLL_EN0							(1 << 7)
#define SI_POLL_EN1							(1 << 6)
#define SI_POLL_EN2							(1 << 5)
#define SI_POLL_EN3							(1 << 4)

#define SI_COMMCSR_TCINT				(1 << 31)
#define SI_COMMCSR_TCINTMSK			(1 << 30)
#define SI_COMMCSR_COMERR				(1 << 29)
#define SI_COMMCSR_RDSTINT			(1 << 28)
#define SI_COMMCSR_RDSTINTMSK		(1 << 27)
#define SI_COMMCSR_TSTART				(1 <<  0)

// write status
#define SI_SR_WRST0							(1 << 28)
#define SI_SR_WRST1							(1 << 20)
#define SI_SR_WRST2							(1 << 12)
#define SI_SR_WRST3							(1 <<  4)
// read status
#define SI_SR_RDST0							(1 << 29)
#define SI_SR_RDST1							(1 << 21)
#define SI_SR_RDST2							(1 << 13)
#define SI_SR_RDST3							(1 <<  5)


#define BUTTON_LEFT			0x0001
#define BUTTON_RIGHT		0x0002
#define BUTTON_DOWN			0x0004
#define BUTTON_UP				0x0008
#define BUTTON_Z				0x0010
#define BUTTON_L				0x0020
#define BUTTON_R				0x0040
#define BUTTON_A				0x0100
#define BUTTON_B				0x0200
#define BUTTON_X				0x0400
#define BUTTON_Y				0x0800
#define BUTTON_START		0x1000

#define ACTION_BUTTON		0x00000000
#define ACTION_AXIS			0x00002000
#define ACTION_AXIS0M		(ACTION_AXIS + 0)
#define ACTION_AXIS0P		(ACTION_AXIS + 1)
#define ACTION_AXIS1M		(ACTION_AXIS + 2)
#define ACTION_AXIS1P		(ACTION_AXIS + 3)
#define ACTION_AXIS2M		(ACTION_AXIS + 4)
#define ACTION_AXIS2P		(ACTION_AXIS + 5)
#define ACTION_AXIS3M		(ACTION_AXIS + 6)
#define ACTION_AXIS3P		(ACTION_AXIS + 7)


void si_init (void);
void si_poll (void);



void pad_set_action (int n, __u16 buttons, __s8 xstick, __s8 ystick,
										 __s8 cxstick, __s8 cystick, __s8 trigl, __s8 trigr);
void pad_action_buttons (int n, int pressed, __u16 buttons);
void pad_action_buttons_masked (int n, int pressed, __u16 buttons, int mask);
void pad_action_analogs (int n, __s8 x, __s8 y);
void pad_action_canalogs (int n, __s8 x, __s8 y);
void pad_action_triggers (int n, __s8 l, __s8 r);
void pad_action_axis (int n, int axis, __s8 value);
void pad_action (int n, int pressed, int action);


#endif // __HW_SI_H
