/*
 *  gcube
 *  Copyright (c) 2004 monk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *  VI - video interface (0xcc002000)
 *      
 *         
 */

#include "hw_vi.h"


__u8 rvi[RVI_SIZE] = {0};


__u32 vi_r32_direct (__u32 addr)
{
	DEBUG (EVENT_LOG, "..vi: read  [%.4x] (%.8x)", addr & 0xffff, RVI32 (addr));
	return RVI32 (addr);
}


void vi_w32_direct (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG, "..vi: write [%.4x] (%.8x) = %.8x", addr & 0xffff, RVI32 (addr), data);
	RVI32 (addr) = data;
}


__u16 vi_r16_direct (__u32 addr)
{
	DEBUG (EVENT_LOG, "..vi: read  [%.4x] (%.4x)", addr & 0xffff, RVI16 (addr));
	return RVI16 (addr);
}


void vi_w16_direct (__u32 addr, __u16 data)
{
	DEBUG (EVENT_LOG, "..vi: write [%.4x] (%.4x) = %.4x", addr & 0xffff, RVI16 (addr), data);
	RVI16 (addr) = data;
}

// !set correct num of lines
void vi_w16_mode (__u32 addr, __u16 data)
{
	DEBUG (EVENT_INFO, "..vi: setting resolution %d / %d (%4.4x)", 640, 480, data);
	video_init (640, 480);

	RVI16 (addr) = data;
}


void vi_w32_mode (__u32 addr, __u32 data)
{
	RVI32 (addr) = data;
	vi_w16_mode (addr, (__u16) data);
}


// external framebuffer half one
void vi_w32_xfb1 (__u32 addr, __u32 data)
{
	DEBUG (EVENT_INFO, "..vi: xfb 1 address set to 0x%.8x", data);

	video_set_framebuffer (MEM_ADDRESS (data));

	RVI32 (addr) = data;
}


// external framebuffer half two
void vi_w32_xfb2 (__u32 addr, __u32 data)
{
	DEBUG (EVENT_INFO, "..vi: xfb 2 address set to 0x%.8x", data);
	RVI32 (addr) = data;
}


void vi_w16_xfb1 (__u32 addr, __u16 data)
{
	RVI16 (addr) = data;
	DEBUG (EVENT_INFO, "..vi: xfb 1 address set to 0x%.8x", RVI32 (0x201c));

	video_set_framebuffer (MEM_ADDRESS (RVI32 (0x201c)));
}


void vi_w16_xfb2 (__u32 addr, __u16 data)
{
	RVI16 (addr) = data;
	DEBUG (EVENT_INFO, "..vi: xfb 2 address set to 0x%.8x", RVI32 (0x2024));
}


// get current vertical position
__u16 vi_r16_vct (__u32 addr)
{
	// fake vsync
	if (RVI16 (VI_VCT) == 0)
	{
		RVI16 (VI_VCT) = 1;
		return 0;
	}
	else if (RVI16 (VI_VCT) == 1)
		return (RVI16 (VI_VCT) = 200);
	else if (RVI16 (VI_VCT) == 200)
		return (RVI16 (VI_VCT) = 522);
	else if (RVI16 (VI_VCT) == 522)
		return (RVI16 (VI_VCT) = 574);
	else
	{
		gcube_refresh_manual ();
		return (RVI16 (VI_VCT) = 0);
	}
}


// current horizontal position of the beam
__u16 vi_r16_hct (__u32 addr)
{
	return (RVI16 (VI_HCT) ^= 639);
}


void vi_w32_viint (__u32 addr, __u32 data)
{
	DEBUG (EVENT_LOG, "..vi: display interrupt [%.4x] set to %.8x", addr & 0xffff, data);
	RVI32 (addr) = data;
}


void vi_w16_viint (__u32 addr, __u16 data)
{
	DEBUG (EVENT_LOG, "..vi: display interrupt [%.4x] set to %.4x", addr & 0xffff, data);
	RVI16 (addr) = data;
}


void vi_refresh (void)
{

	if (RVI32 (VI_DINT0) & VI_DI_INTEN)
		RVI32 (VI_DINT0) |= VI_DI_INT;

	if (RVI32 (VI_DINT1) & VI_DI_INTEN)
		RVI32 (VI_DINT1) |= VI_DI_INT;

	if (RVI32 (VI_DINT2) & VI_DI_INTEN)
		RVI32 (VI_DINT2) |= VI_DI_INT;

	if (RVI32 (VI_DINT3) & VI_DI_INTEN)
		RVI32 (VI_DINT3) |= VI_DI_INT;


	video_refresh ();
	si_poll ();

	if ((RVI32 (VI_DINT0) |
			 RVI32 (VI_DINT1) |
			 RVI32 (VI_DINT2) |
			 RVI32 (VI_DINT3)) & VI_DI_INTEN)
		pi_interrupt (INTERRUPT_VI);
}


// for savestates
void vi_reinit_video (void)
{
	// video init
	if (RVI32 (VI_XFB1))
		vi_w32_xfb1 (VI_XFB1, RVI32 (VI_XFB1));
	else if (RVI32 (VI_MODE))
		vi_w32_mode (VI_MODE, RVI32 (VI_MODE));
}


void vi_init (void)
{
	memset (rvi, 0, sizeof (rvi));

	// video mode
	mem_hww_hook (16, 0x2000, vi_w16_mode);
	mem_hwr_hook (16, 0x2000, vi_r16_direct);
	mem_hww_hook (16, 0x2002, vi_w16_mode);
	mem_hwr_hook (16, 0x2002, vi_r16_direct);
	mem_hww_hook (32, 0x2000, vi_w32_mode);
	mem_hwr_hook (32, 0x2000, vi_r32_direct);

	// framebuffer
	mem_hww_hook (32, 0x201c, vi_w32_xfb1);
	mem_hww_hook (32, 0x2024, vi_w32_xfb2);
	mem_hwr_hook (32, 0x201c, vi_r32_direct);
	mem_hwr_hook (32, 0x2024, vi_r32_direct);

	mem_hww_hook (16, 0x201c, vi_w16_direct);
	mem_hww_hook (16, 0x201e, vi_w16_xfb1);
	mem_hww_hook (16, 0x2024, vi_w16_direct);
	mem_hww_hook (16, 0x2026, vi_w16_xfb2);

	// current position of the beam
	mem_hwr_hook (16, 0x202c, vi_r16_vct);
	mem_hwr_hook (16, 0x202e, vi_r16_hct);
	mem_hww_hook (32, 0x202c, mem_fake_w32);

	// display interrupt 0 1 2 3
	mem_hwr_hook (32, 0x2030, vi_r32_direct);
	mem_hwr_hook (32, 0x2034, vi_r32_direct);
	mem_hwr_hook (32, 0x2038, vi_r32_direct);
	mem_hwr_hook (32, 0x203c, vi_r32_direct);
	mem_hww_hook (32, 0x2030, vi_w32_viint);
	mem_hww_hook (32, 0x2034, vi_w32_viint);
	mem_hww_hook (32, 0x2038, vi_w32_viint);
	mem_hww_hook (32, 0x203c, vi_w32_viint);

	mem_hwr_hook (16, 0x2030, vi_r16_direct);
	mem_hwr_hook (16, 0x2032, vi_r16_direct);
	mem_hwr_hook (16, 0x2034, vi_r16_direct);
	mem_hwr_hook (16, 0x2036, vi_r16_direct);
	mem_hwr_hook (16, 0x2038, vi_r16_direct);
	mem_hwr_hook (16, 0x203a, vi_r16_direct);
	mem_hwr_hook (16, 0x203c, vi_r16_direct);
	mem_hwr_hook (16, 0x203e, vi_r16_direct);

	mem_hww_hook (16, 0x2030, vi_w16_viint);
	mem_hww_hook (16, 0x2032, vi_w16_viint);
	mem_hww_hook (16, 0x2034, vi_w16_viint);
	mem_hww_hook (16, 0x2036, vi_w16_viint);
	mem_hww_hook (16, 0x2038, vi_w16_viint);
	mem_hww_hook (16, 0x203a, vi_w16_viint);
	mem_hww_hook (16, 0x203c, vi_w16_viint);
	mem_hww_hook (16, 0x203e, vi_w16_viint);

	// ignored

	// timings
	mem_hww_hook (32, 0x2004, mem_fake_w32);
	mem_hww_hook (32, 0x2008, mem_fake_w32);
	mem_hww_hook (32, 0x200c, mem_fake_w32);
	mem_hww_hook (32, 0x2010, mem_fake_w32);
	mem_hww_hook (32, 0x2014, mem_fake_w32);
	mem_hww_hook (32, 0x2018, mem_fake_w32);

	mem_hww_hook (16, 0x2002, mem_fake_w16);
	mem_hww_hook (16, 0x2004, mem_fake_w16);
	mem_hww_hook (16, 0x2006, mem_fake_w16);
	mem_hww_hook (16, 0x2008, mem_fake_w16);
	mem_hww_hook (16, 0x200a, mem_fake_w16);
	mem_hww_hook (16, 0x200c, mem_fake_w16);
	mem_hww_hook (16, 0x200e, mem_fake_w16);
	mem_hww_hook (16, 0x2010, mem_fake_w16);
	mem_hww_hook (16, 0x2012, mem_fake_w16);
	mem_hww_hook (16, 0x2014, mem_fake_w16);
	mem_hww_hook (16, 0x2016, mem_fake_w16);
	mem_hww_hook (16, 0x2018, mem_fake_w16);
	mem_hww_hook (16, 0x201a, mem_fake_w16);

	// top and bottom field base registers (for 3d mode)
	mem_hww_hook (32, 0x2020, mem_fake_w32);
	mem_hww_hook (32, 0x2028, mem_fake_w32);

	// scaling register
	mem_hww_hook (16, 0x2048, mem_fake_w16);
	mem_hww_hook (16, 0x204a, mem_fake_w16);
	mem_hww_hook (32, 0x2048, mem_fake_w32);

	// filtering registers (anti-aliasing)
	mem_hww_hook (32, 0x204c, mem_fake_w32);
	mem_hww_hook (32, 0x2050, mem_fake_w32);
	mem_hww_hook (32, 0x2054, mem_fake_w32);
	mem_hww_hook (32, 0x2058, mem_fake_w32);
	mem_hww_hook (32, 0x205c, mem_fake_w32);
	mem_hww_hook (32, 0x2060, mem_fake_w32);
	mem_hww_hook (32, 0x2064, mem_fake_w32);
	mem_hww_hook (32, 0x2068, mem_fake_w32);

	mem_hww_hook (16, 0x204c, mem_fake_w16);
	mem_hww_hook (16, 0x204e, mem_fake_w16);
	mem_hww_hook (16, 0x2050, mem_fake_w16);
	mem_hww_hook (16, 0x2052, mem_fake_w16);
	mem_hww_hook (16, 0x2054, mem_fake_w16);
	mem_hww_hook (16, 0x2056, mem_fake_w16);
	mem_hww_hook (16, 0x2058, mem_fake_w16);
	mem_hww_hook (16, 0x205a, mem_fake_w16);
	mem_hww_hook (16, 0x205c, mem_fake_w16);
	mem_hww_hook (16, 0x205e, mem_fake_w16);
	mem_hww_hook (16, 0x2060, mem_fake_w16);
	mem_hww_hook (16, 0x2062, mem_fake_w16);
	mem_hww_hook (16, 0x2064, mem_fake_w16);
	mem_hww_hook (16, 0x2066, mem_fake_w16);


	// vi clock select / dtv status register
	mem_hwr_hook (16, 0x206c, mem_fake_r16);
	mem_hww_hook (16, 0x206c, mem_fake_w16);
	mem_hww_hook (32, 0x206c, mem_fake_w32);

	// border
	mem_hww_hook (32, 0x2070, mem_fake_w32);
	mem_hww_hook (32, 0x2074, mem_fake_w32);

	mem_hww_hook (16, 0x2070, mem_fake_w16);
	mem_hww_hook (16, 0x2072, mem_fake_w16);
	mem_hww_hook (16, 0x2074, mem_fake_w16);
	mem_hww_hook (16, 0x2076, mem_fake_w16);

	// unknown
	mem_hww_hook (32, 0x2040, mem_fake_w32);
	mem_hww_hook (32, 0x2044, mem_fake_w32);
	mem_hww_hook (32, 0x2078, mem_fake_w32);
	mem_hww_hook (32, 0x207c, mem_fake_w32);
}
