/*
 *  gcube
 *  Copyright (c) 2004 monk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *
 *      
 *         
 */

#include "mem.h"


#ifdef GDEBUG
# define CHECK_BOUNDS(X)\
	({\
			if ((X & MEM_MASK) > MEM_SIZE)\
				DEBUG (EVENT_EFATAL, ".mem: access outside of mem boundary 0x%.8x", X & MEM_MASK);\
	})
#else
# define CHECK_BOUNDS(X)
#endif

#define MEMORY_READ(X)		CHECK_BOUNDS(X)
#define MEMORY_WRITE(X)		CHECK_BOUNDS(X)

#define UNHANDLED_READ(AC,X)		(DEBUG (EVENT_UHW_READ, ".mem: unhandled hw read%d (0x%.8x)", AC, addr))
#define UNHANDLED_WRITE(AC,X,d)	(DEBUG (EVENT_UHW_WRITE, ".mem: unhandled hw write%d (0x%.8x) = %.8x", AC, addr, d))



// memory
#ifdef MEMHACK
__u8 RAM[2][MEM_SIZE] = {{0,0}};
#else
__u8 RAM[MEM_SIZE] = {0};
#endif


// hardware registers hooks
__u8  (*hwread8   [0xffff]) (__u32 addr);
__u16 (*hwread16  [0xffff]) (__u32 addr);
__u32 (*hwread32  [0xffff]) (__u32 addr);
void  (*hwwrite8  [0xffff]) (__u32 addr, __u8 data);
void  (*hwwrite16 [0xffff]) (__u32 addr, __u16 data);
void  (*hwwrite32 [0xffff]) (__u32 addr, __u32 data);




// reads

inline __u8 read_byte (__u32 addr)
{
	if (addr > 0xcc000000)
		return hwread8 [addr & 0xffff] (addr);
	else
	{
		MEMORY_READ (addr);
		return (MEM (addr));
	}
}


inline __u16 read_half_word_r (__u32 addr)
{
	MEMORY_READ (addr);
	return MEM16 (addr);
}


inline __u16 read_half_word (__u32 addr)
{
	if (addr > 0xcc000000)
		return hwread16 [addr & 0xfffe] (addr);
	else
		return BSWAP16 (read_half_word_r (addr));
}


inline __u32 read_word_r (__u32 addr)
{
	MEMORY_READ (addr);
	return MEM32 (addr);
}


inline __u32 read_word (__u32 addr)
{
	if (addr > 0xcc000000)
		return hwread32 [addr & 0xfffc] (addr);
	else
		return BSWAP32 (read_word_r (addr));
}


inline __u64 read_double_r (__u32 addr)
{
	MEMORY_READ (addr);
	return MEM64 (addr);
}


inline __u64 read_double (__u32 addr)
{
	return BSWAP64 (read_double_r (addr));
}


// writes

inline void write_byte (__u32 addr, __u8 data)
{
	if (addr > 0xcc000000)
		return hwwrite8 [addr & 0xffff] (addr, data);

	MEMORY_WRITE (addr);
	MEM (addr) = data;
}


inline void write_half_word_r (__u32 addr, __u16 data)
{
	MEMORY_WRITE (addr);
	MEM16 (addr) = data;
}


inline void write_half_word (__u32 addr, __u16 data)
{
	if (addr > 0xcc000000)
		return hwwrite16 [addr & 0xfffe] (addr, data);

	MEMORY_WRITE (addr);
	MEM16 (addr) = BSWAP16 (data);
}


inline void write_word_r (__u32 addr, __u32 data)
{
	MEMORY_WRITE (addr);
	MEM32 (addr) = data;
}


inline void write_word (__u32 addr, __u32 data)
{
	if (addr > 0xcc000000)
		return hwwrite32 [addr & 0xfffc] (addr, data);

	MEMORY_WRITE (addr);
	MEM32 (addr) = BSWAP32 (data);
}


inline void write_double (__u32 addr, __u64 data)
{
	MEMORY_WRITE (addr);
	MEM64 (addr) = BSWAP64 (data);
}


// hw registers handling

void mem_fake_w32 (__u32 addr, __u32 data)
{
}


__u32 mem_fake_r32 (__u32 addr)
{
	return 0;
}


void mem_fake_w16 (__u32 addr, __u16 data)
{
}


__u16 mem_fake_r16 (__u32 addr)
{
	return 0;
}


// hooks

void mem_hw_hook (int read, int bits, __u32 addr, void *ptr)
{
	switch (bits)
	{
		case 8:
			if (read)
				hwread8[addr & 0xffff] = ptr;
			else
				hwwrite8[addr & 0xffff] = ptr;
			break;

		case 16:
			if (read)
				hwread16[addr & 0xffff] = ptr;
			else
				hwwrite16[addr & 0xffff] = ptr;
			break;

		case 32:
			if (read)
				hwread32[addr & 0xffff] = ptr;
			else
				hwwrite32[addr & 0xffff] = ptr;
			break;
	}
}


// hardware read hook
void mem_hwr_hook (int bits, __u32 addr, void *ptr)
{
	mem_hw_hook (TRUE, bits, addr, ptr);
}


// hardware write hook
void mem_hww_hook (int bits, __u32 addr, void *ptr)
{
	mem_hw_hook (FALSE, bits, addr, ptr);
}


__u8 hook_unhandled_read8 (__u32 addr)
{
	UNHANDLED_READ (8, addr & 0xffff);
	return 0;
}


void hook_unhandled_write8 (__u32 addr, __u8 data)
{
	UNHANDLED_WRITE (8, addr & 0xffff, data);
}


__u16 hook_unhandled_read16 (__u32 addr)
{
	UNHANDLED_READ (16, addr & 0xffff);
	return 0;
}


void hook_unhandled_write16 (__u32 addr, __u16 data)
{
	UNHANDLED_WRITE (16, addr & 0xffff, data);
}


__u32 hook_unhandled_read32 (__u32 addr)
{
	UNHANDLED_READ (32, addr & 0xffff);
	return 0;
}


void hook_unhandled_write32 (__u32 addr, __u32 data)
{
	UNHANDLED_WRITE (32, addr & 0xffff, data);
}


void mem_reset (void)
{
	int i;


	memset (RAM, 0, MEM_SIZE);

	for (i = 0; i < 0xffff; i++)
	{
		mem_hwr_hook (8,  i, hook_unhandled_read8);
		mem_hwr_hook (16, i, hook_unhandled_read16);
		mem_hwr_hook (32, i, hook_unhandled_read32);
		mem_hww_hook (8,  i, hook_unhandled_write8);
		mem_hww_hook (16, i, hook_unhandled_write16);
		mem_hww_hook (32, i, hook_unhandled_write32);
	}

	DEBUG (EVENT_INFO, ".mem: reinitialized");
}
