#ifndef __MEM_H
#define __MEM_H 1


#include <stdio.h>
#include <string.h>

#include "general.h"
#include "gcube.h"
#include "cpu.h"
#include "hw_ai_dsp.h"
#include "hw_cp.h"
#include "hw_di.h"
#include "hw_exi.h"
#include "hw_gx.h"
#include "hw_mi.h"
#include "hw_pe.h"
#include "hw_pi.h"
#include "hw_si.h"
#include "hw_vi.h"

// memory
#define MEM_SIZE 0x01800000
#define MEM_MASK 0x01ffffff

// some programs will set the fb too low and overwrite the code
// this temporary hack gets around that
//#define MEMHACK

#ifdef MEMHACK
extern __u8 RAM[2][MEM_SIZE];
# define MEM(X)					(RAM[((X) >> 30) & 1][(X) & MEM_MASK])
# define MEM_SAFE(X)		(RAM[((X) >> 30) & 1][(((X) & MEM_MASK) > MEM_SIZE) ? 0 : ((X) & MEM_MASK)])
#else
extern __u8 RAM[MEM_SIZE];
# define MEM(X)					(RAM[(X) & MEM_MASK])
# define MEM_SAFE(X)		(RAM[(((X) & MEM_MASK) > MEM_SIZE) ? 0 : ((X) & MEM_MASK)])
#endif

#define MEM32_SAFE(A)		(*((__u32 *) &MEM_SAFE (A)))
#define MEM_ADDRESS(X)	(&MEM (X))


typedef union
{
	__u64 bin;
	double fp;
} Double64;


typedef union
{
	__u32 bin;
	float fp;
} Single32;


#define MEM16(A)			(*((__u16 *) &MEM (A)))
#define MEM32(A)			(*((__u32 *) &MEM (A)))
#define MEM64(A)			(*((__u64 *) &MEM (A)))

#define MEMS16(A)			(*((__s16 *) &MEM (A)))
#define MEMS32(A)			(*((__s32 *) &MEM (A)))
#define MEMS64(A)			(*((__s64 *) &MEM (A)))

#define MEMF(A)				(*((float *) &MEM (A)))
// ugly
#define MEMRF(A)\
	({\
		__u32 d = MEMR32 (A);\
		float *f = (float *) &d;\
		*f;\
	})

#define MEMR16(A)			(BSWAP16 (MEM16 (A)))
#define MEMR32(A)			(BSWAP32 (MEM32 (A)))
#define MEMR64(A) 		(BSWAP64 (MEM64 (A)))

#define MEMSR16(A)		((__s16) BSWAP16 (MEM16 (A)))
#define MEMSR32(A)		((__s32) BSWAP32 (MEM32 (A)))
#define MEMSR64(A)		((__s64) BSWAP64 (MEM64 (A)))

#define MEMW(A,d)			(MEM (A) = d)
#define MEMWR16(A,d)	(MEM16 (A) = BSWAP16 (d))
#define MEMWR32(A,d)	(MEM32 (A) = BSWAP32 (d))


// byte swap
#ifndef BSWAP16
# define BSWAP16(B)\
	({\
			__u16 _B = (B);\
			(((__u16) ((_B) << 8) | ((_B) >> 8)));\
	})
#endif

#ifndef BSWAP32
# define BSWAP32(B)\
	({\
			__u32 _B = (B);\
			(((__u32) (_B << 24) | ((_B << 8) & 0xff0000) | ((_B >> 8) & 0xff00) | (_B >> 24)));\
	})
#endif

#ifndef BSWAP64
# define BSWAP64(B)\
	({\
			__u64 _B = (B);\
			__u32 *_B32 = (__u32 *) &_B;\
			(((__u64) BSWAP32 (_B32[0])) << 32) | BSWAP32 (_B32[1]);\
	})
#endif

// reads
#define MEM_RBYTE				read_byte
#define MEM_RHALF				read_half_word
#define MEM_RHALF_S			(__s16) read_half_word
#define MEM_RHALF_SR		(__s16) read_half_word_r
#define MEM_RWORD				read_word
#define MEM_RWORD_R			read_word_r

// writes
#define MEM_WBYTE				write_byte
#define MEM_WHALF				write_half_word
#define MEM_WHALF_R			write_half_word_r
#define MEM_WWORD				write_word
#define MEM_WWORD_R			write_word_r

#define MEM_WSINGLE(A,d)			(write_word (A, ((Single32) (d)).bin))
#define MEM_RSINGLE_F(A)			(((Single32) read_word (A)).fp)
#define MEM_RSINGLE_B(A)			(((Single32) read_word (A)).bin)

#define MEM_WDOUBLE(A,d)			(write_double (A, ((Double64) (d)).bin))
#define MEM_RDOUBLE_F(A)			(((Double64) read_double (A)).fp)
#define MEM_RDOUBLE_B(A)			(((Double64) read_double (A)).bin)


// memory map
#define MEM_MAGIC								0x80000020
#define MEM_VERSION							0x80000024
#define MEM_MEM_SIZE						0x80000028
#define MEM_CONSOLE_TYPE				0x8000002c
#define MEM_HEAP_BOTTOM					0x80000030
#define MEM_HEAP_TOP						0x80000034
#define MEM_FST									0x80000038
#define MEM_FST_SIZE						0x8000003c
#define MEM_DEBUG								0x80000040
#define MEM_SIM_MEM_SIZE				0x800000f0
#define MEM_BUS_CLOCK_SPEED			0x800000f8
#define MEM_CPU_CLOCK_SPEED			0x800000fc

#define CONSOLE_TYPE_RETAIL1		1
#define CONSOLE_TYPE_HW2_PB			2
#define CONSOLE_TYPE_LATEST_PB	3

#define MEM_MAGIC_NORMAL				0x0D15EA5E
#define MEM_MAGIC_JTAG					0xE5207C22
// this really depends on the application
#define ARENA_LO								0x80403100
#define ARENA_HI								0x81600000



void mem_reset (void);

__u8 read_byte (__u32 addr);
__u16 read_half_word_r (__u32 addr);
__u16 read_half_word (__u32 addr);
__u32 read_word_r (__u32 addr);
__u32 read_word (__u32 addr);
__u64 read_double (__u32 addr);
void write_byte (__u32 addr, __u8 data);
void write_half_word_r (__u32 addr, __u16 data);
void write_half_word (__u32 addr, __u16 data);
void write_word_r (__u32 addr, __u32 data);
void write_word (__u32 addr, __u32 data);
void write_double (__u32 addr, __u64 data);

void mem_fake_w32 (__u32 addr, __u32 data);
__u32 mem_fake_r32 (__u32 addr);
void mem_fake_w16 (__u32 addr, __u16 data);
__u16 mem_fake_r16 (__u32 addr);

void mem_hwr_hook (int bits, __u32 addr, void *ptr);
void mem_hww_hook (int bits, __u32 addr, void *ptr);


#endif // __MEM_H
