/*
 *  gcube
 *  Copyright (c) 2004 monk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 *
 *      
 *         
 */

#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

#include "video.h"


SDL_Surface *offscreen = NULL;
unsigned int fbtex = 0, fbdl = 0;

SDL_Surface *screen = NULL;
SDL_Overlay *yuv = NULL;
SDL_Rect r;
SDL_Joystick *gamepad[4] = {0};
int ngamepads = 0;

unsigned int display_mode = 0;

int display_fb = FALSE;
int wireframe = FALSE;
extern int use_textures, use_colors;



#include "icon.c"
void video_set_icon (void)
{
	SDL_Surface *ic;


	ic = SDL_CreateRGBSurfaceFrom ((void *)icon.pixel_data, icon.width,
				icon.height, icon.bytes_per_pixel*8, icon.width * icon.bytes_per_pixel,
				0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000);
											
	SDL_WM_SetIcon (ic, NULL);
}


// reinitialize if screen is not NULL
int video_init (int w, int h)
{
	if ((screen && ((screen->w == w) && (screen->h == h))))
		return FALSE;

	if (!screen)
	{
		if (0 > SDL_Init (SDL_INIT_VIDEO | SDL_INIT_JOYSTICK | SDL_INIT_AUDIO))
			return FALSE;
		atexit (video_close);
	}

	video_set_icon ();
	screen = SDL_SetVideoMode (w, h, VID_BPP, SDL_HWSURFACE | SDL_OPENGL);

	// check for extensions
	if (!strstr (glGetString (GL_EXTENSIONS), "_texture_rectangle"))
		gcube_quit ("Textured rectangle extension is not supported by Your video card!");
	
	if (offscreen)
		SDL_FreeSurface (offscreen);
	offscreen = SDL_CreateRGBSurface (SDL_HWSURFACE, w, h, VID_BPP,
																		0x00ff0000, 0x0000ff00, 0x000000ff, 0xff000000);

	if (yuv)
	{
		void *pixels = yuv->pixels[0];

		SDL_FreeYUVOverlay (yuv);
		yuv = SDL_CreateYUVOverlay (w, h, SDL_YUY2_OVERLAY, offscreen);
		yuv->pixels[0] = pixels;
	}
	else
		yuv = SDL_CreateYUVOverlay (w, h, SDL_YUY2_OVERLAY, offscreen);

	if (GL_TEXTURE_RECT == GL_TEXTURE_2D)
	{
		glMatrixMode (GL_TEXTURE);
		glLoadIdentity ();
		glScalef (1.0/w, 1.0/h, 1);
	}

	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	glViewport (0, 0, w, h);
	glOrtho (0, w, h, 0, -1, 1);
	glClearColor (1, 1, 1, 0);
	glClear (GL_COLOR_BUFFER_BIT);

	glEnable (GL_TEXTURE_RECT);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	if (fbtex)
		glDeleteTextures (1, &fbtex);
	glGenTextures (1, &fbtex);
	glBindTexture (GL_TEXTURE_RECT, fbtex);
	glTexParameteri (GL_TEXTURE_RECT, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri (GL_TEXTURE_RECT, GL_TEXTURE_WRAP_T, GL_CLAMP);
	// those two are necessery
	glTexParameteri (GL_TEXTURE_RECT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_RECT, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D (GL_TEXTURE_RECT, 0, GL_RGB, offscreen->w, offscreen->h, 0, GL_RGB,
								GL_UNSIGNED_BYTE, offscreen->pixels);

	if (fbdl)
		glDeleteLists (fbdl, 1);
	fbdl = glGenLists (1);
	glNewList (fbdl, GL_COMPILE);

		glEnable (GL_TEXTURE_RECT);
		glBegin (GL_QUADS);
			glTexCoord2f (0, h);
			glVertex2f (0, h);

			glTexCoord2f (w, h);
			glVertex2f (w, h);

			glTexCoord2f (w, 0);
			glVertex2f (w, 0);

			glTexCoord2f (0, 0);
			glVertex2f (0, 0);
		glEnd ();
		glDisable (GL_TEXTURE_RECT);

	glEndList ();

	r.x = r.y = 0;
	r.w = w;
	r.h = h;

	DEBUG (EVENT_INFO, ".input: %d gamepad(s) connected", SDL_NumJoysticks ());
	if (SDL_NumJoysticks ())
	{
		int i;
	
		for (i = 0; i < 4; i++)
			if (gamepad[i])
				SDL_JoystickClose (gamepad[i]);

		SDL_JoystickEventState (SDL_ENABLE);

		ngamepads = SDL_NumJoysticks ();
		if (ngamepads > 4)
			ngamepads = 4;		

		for (i = 0; i < ngamepads; i++)
		{
			gamepad[i] = SDL_JoystickOpen (i);
		
			DEBUG (EVENT_INFO, "        [%s] with %d buttons, %d axes, %d hats and %d balls",
							SDL_JoystickName (i), SDL_JoystickNumButtons (gamepad[i]),
							SDL_JoystickNumAxes (gamepad[i]), SDL_JoystickNumHats (gamepad[i]),
							SDL_JoystickNumBalls (gamepad[i]));
		}
	}
	
	SDL_WM_SetCaption (GCUBE_DESCRIPTION, NULL);

	audio_init (FREQ_32KHZ, CONFIG_AUDIO_BUFFER_SIZE);

	return TRUE;
}


void video_set_title (char *title)
{
	char buff[256];


	sprintf (buff, "%s %s", GCUBE_DESCRIPTION, title);
	SDL_WM_SetCaption (buff, NULL);
}


void video_close (void)
{
	int i;
	

	if (!screen)
		return;

	if (yuv)
	{
		SDL_FreeYUVOverlay (yuv);
		yuv = NULL;
	}

	for (i = 0; i < 4; i++)
		if (gamepad[i])
		{
			SDL_JoystickClose (gamepad[i]);
			gamepad[i] = NULL;
		}

	glDeleteTextures (1, &fbtex);
	glDeleteLists (fbdl, 1);

	screen = NULL;

	SDL_Quit ();
}


void video_draw (void)
{
	if (!yuv)
		return;

	if (display_fb)
	{
		SDL_DisplayYUVOverlay (yuv, &r);

		glBindTexture (GL_TEXTURE_RECT, fbtex);
		glTexSubImage2D (GL_TEXTURE_RECT, 0, 0, 0, offscreen->w, offscreen->h,
										 GL_BGR, GL_UNSIGNED_BYTE, offscreen->pixels);
	}

	SDL_GL_SwapBuffers ();

	if (wireframe)
		glClear (GL_COLOR_BUFFER_BIT);

	if (display_fb)
		glCallList (fbdl);
}


void input_check (void)
{
	SDL_Event event;
	Configuration *config = config_get ();
	static const int buttons[] =
	{
		BUTTON_START, BUTTON_A, BUTTON_B, BUTTON_L, BUTTON_R,
		BUTTON_X, BUTTON_Y, BUTTON_Z, BUTTON_UP, BUTTON_DOWN,
		BUTTON_LEFT, BUTTON_RIGHT, ACTION_AXIS0M, ACTION_AXIS0P,
		ACTION_AXIS1M, ACTION_AXIS1P, ACTION_AXIS2M, ACTION_AXIS2P,
		ACTION_AXIS3M, ACTION_AXIS3P,
	};


	if (SDL_PollEvent (&event))
	{
		switch (event.type)
		{
			case SDL_QUIT:
				gcube_quit ("quit");
				break;

			case SDL_KEYDOWN:
				switch (event.key.keysym.sym)
				{
					case SDLK_F2:
						wireframe = !wireframe;
						if (wireframe)
						{
							glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
							glDisable (GL_TEXTURE_RECT);
							glColor4f (0, 0, 0, 1);
							use_textures = FALSE;
							use_colors = FALSE;
							
							video_set_title ("(wireframe)");
						}
						else
						{
							glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
							glColor4f (1, 1, 1, 1);
							use_textures = TRUE;
							use_colors = TRUE;
							
							video_set_title ("");
						}
						break;

					case SDLK_F10:
						SDL_WM_ToggleFullScreen (screen);
						break;

					case SDLK_F3:
						if (display_mode != 574)
							video_init (640, (display_mode = 574));
						else
							video_init (640, (display_mode = 480));
						break;

					case SDLK_c:
						if (event.key.keysym.mod == KMOD_LCTRL || event.key.keysym.mod == KMOD_RCTRL)
							DEBUG (EVENT_STOP, ".video: user interrupt");
						break;
						
					default:
						break;
				}
				// fall through
				
			case SDL_KEYUP:
				switch (event.key.keysym.sym)
				{
					case SDLK_ESCAPE:
						gcube_quit ("quit");
						break;

					default:
						{
							int i, j;
						
							for (j = 0; j < 4; j++)
								if (config->keyboard[j])
									for (i = 0; i < sizeof (buttons) / sizeof (*buttons); i++)
										if (event.key.keysym.sym == config->kbmap[j][i])
											pad_action (config->keyboard[j] - 1,
																	(event.type == SDL_KEYDOWN), buttons[i]);
						}
						break;
				}
				break;

			case SDL_JOYBUTTONDOWN:
			case SDL_JOYBUTTONUP:
				{
					int i;
						
					if (config->gamepad[event.jbutton.which])
					{
						// only 8 buttons for gamepad (no action_axis)
						for (i = 0; i < 8; i++)
							if (config->padbmap[event.jbutton.which][i] == event.jbutton.button + 1)
								pad_action (config->gamepad[event.jbutton.which] - 1,
														(event.type == SDL_JOYBUTTONDOWN), buttons[i]);
					}
				}
				break;
				
			case SDL_JOYHATMOTION:
				{
					unsigned int buttons = 0;

					if (!config->gamepad[event.jbutton.which])
						break;

					if (event.jhat.value & SDL_HAT_UP)
						buttons |= BUTTON_UP;

					if (event.jhat.value & SDL_HAT_DOWN)
						buttons |= BUTTON_DOWN;

					if (event.jhat.value & SDL_HAT_LEFT)
						buttons |= BUTTON_LEFT;

					if (event.jhat.value & SDL_HAT_RIGHT)
						buttons |= BUTTON_RIGHT;

					pad_action_buttons_masked (config->gamepad[event.jbutton.which] - 1,
																		TRUE,	buttons,
																		BUTTON_UP | BUTTON_DOWN | BUTTON_LEFT | BUTTON_RIGHT);
				}
				break;
				
			case SDL_JOYAXISMOTION:
				if (config->gamepad[event.jaxis.which])
				{
					int i;
				
					for (i = 0; i < 4; i++)
						if (config->padamap[event.jaxis.which][i] == event.jaxis.axis + 1)
							pad_action_axis (config->gamepad[event.jaxis.which] - 1,
															 i, event.jaxis.value >> 8);
				}
				break;
		}
	}
}


void video_refresh (void)
{
	if (screen)
	{
		video_draw ();
		input_check ();
	}
}


void video_set_framebuffer (unsigned char *addr)
{
	if (!yuv && !video_init (640, 480))
		return;
		
	SDL_LockYUVOverlay (yuv);
	yuv->pixels[0] = addr;
	SDL_UnlockYUVOverlay (yuv);
	
	display_fb = TRUE;
}


void video_dump_texture (char *filename, char *data, int w, int h, int format)
{
	SDL_Surface *surf;


	surf = SDL_CreateRGBSurfaceFrom (gx_convert_texture (data, w, h, format), 
																	 w, h, 32, 4*w,
																	 0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000);
	SDL_SaveBMP (surf, filename);
	SDL_FreeSurface (surf);
}


void video_disable_fb (int disable)
{
	display_fb = !disable;
}


int input_gamepads_connected (void)
{
	return ngamepads;
}
